import unittest
import unittest.mock as mocking
import uuid

from drb.drivers.odata import ODataServiceNodeDias, \
    ODataServiceType, ODataProductNode, ODataServiceNodeList
from drb.drivers.odata.odata_utils import ODataUtils, ODataQueryPredicate
from tests.utils import start_mock_odata_dias, \
    stop_mock_odata_csc, dias_products


class TestODataDiasServiceNode(unittest.TestCase):
    svc_url = 'https://gael-systems.com/odata/dias'
    node = None

    @classmethod
    def setUpClass(cls) -> None:
        start_mock_odata_dias(cls.svc_url)
        cls.node = ODataServiceNodeDias(cls.svc_url)
        cls.node._type = ODataServiceType.ONDA_DIAS

    @classmethod
    def tearDownClass(cls) -> None:
        stop_mock_odata_csc()
        cls.node._type = ODataServiceType.CSC

    def test_name(self):
        self.assertEqual(self.svc_url, self.node.name)

    def test_namespace_uri(self):
        self.assertEqual('Ens', self.node.namespace_uri)

    def test_value(self):
        self.assertIsNone(self.node.value)

    def test_path(self):
        self.assertEqual(self.svc_url, self.node.path.name)

    def test_parent(self):
        self.assertIsNone(self.node.parent)

    def test_attributes(self):
        self.assertEqual({}, self.node.attributes)

    def test_children(self):
        expected = [
            ODataProductNode(
                self.svc_url, 'da4cdfee-9fef-4ab1-b4a2-20f9ec8005bd'),
            ODataProductNode(
                self.svc_url, '34416861-c364-44be-8bc9-16829f32cf31'),
            ODataProductNode(
                self.svc_url, 'c3f89d3d-3a06-4e42-9903-a5d07ab5c642')
        ]
        with mocking.patch.object(ODataUtils, 'req_svc_products',
                                  return_value=(dias_products, -1)):
            children = self.node.children
            self.assertIsNotNone(children)
            self.assertIsInstance(children, list)
            self.assertEqual(expected[1],
                             children[1].get_impl(ODataProductNode))
            self.assertEqual(expected[-1],
                             children[-1].get_impl(ODataProductNode))
            self.assertIsInstance(children[:2], ODataServiceNodeList)
            self.assertIsInstance(children[1:-1], ODataServiceNodeList)

        with mocking.patch.object(ODataUtils, 'req_svc_products',
                                  return_value=dias_products):
            itr = iter(children)
            self.assertEqual(expected[0], next(itr).get_impl(ODataProductNode))
            self.assertEqual(expected[1], next(itr).get_impl(ODataProductNode))
            self.assertEqual(expected[2], next(itr).get_impl(ODataProductNode))

    def test_len(self):
        self.assertEqual(3, len(self.node))

    def test_bracket_browse(self):
        prd2 = {
            'uuid': '34416861-c364-44be-8bc9-16829f32cf31',
            'name': 'S2B_MSIL2A_20180328T120349_N0207_R066_'
                    'T29VLD_20180328T175102.zip'
        }

        # int
        with mocking.patch.object(ODataUtils, 'req_svc_products',
                                  return_value=(dias_products, -1)):
            node = self.node[1]
            self.assertEqual(prd2['uuid'], node.get_attribute('id'))
            node = self.node[-2].get_impl(ODataProductNode)
            self.assertEqual(prd2['uuid'], node.get_attribute('id'))
            with self.assertRaises(IndexError):
                n = self.node[42]

        # # tuple(str, str, int) -> by Name
        actual_node = self.node[prd2['name']].get_impl(ODataProductNode)
        self.assertEqual(prd2['uuid'], actual_node.get_attribute('id'))
        self.assertEqual(prd2['uuid'], actual_node @ 'id')

        n = self.node[ODataQueryPredicate(filter=f"Name eq '{prd2['name']}'")]
        self.assertIsInstance(n, list)
        self.assertEqual(3, len(n))

        with self.assertRaises(KeyError):
            n = self.node['test']
        with self.assertRaises(KeyError):
            n = self.node['test', 1]
        with self.assertRaises(KeyError):
            n = self.node[prd2['name'], 'ns']
        with self.assertRaises(KeyError):
            n = self.node[prd2['name'], 'ns', 1]

        # UUID -> by Id
        self.assertEqual(prd2['name'], self.node[uuid.UUID(prd2['uuid'])].name)
        with self.assertRaises(KeyError):
            fake_uuid = prd2['uuid'][:-2] + 'f'
            n = self.node[fake_uuid]

        with self.assertRaises(TypeError):
            n = self.node[b"helloWorld"]

    def test_equals(self):
        self.assertEqual(self.node, ODataServiceNodeDias(self.svc_url))

    def test_prepare_filter(self):

        result = '%22creationDate:[1900-05-10T00:00:00.000Z TO NOW]%22'
        self.assertEqual(result, self.node.prepare_filter(None))



import unittest

from shapely import Polygon

from drb.drivers.odata import (
    Expression, ExpressionType, ComparisonOperator,
    LogicalOperator, ArithmeticOperator,
    GroupingOperator, ExpressionFunc
)


class TestExpression(unittest.TestCase):
    foot_print = "geography'SRID=4326;Polygon((" \
                 "104.99934004982643 -74.2320605340525," \
                 "105.27272732082909 -74.28534428112921," \
                 "105.25623295629731 -74.291237938206087))'"

    def test_type(self):
        bool = ExpressionType.bool(True)
        self.assertIsNotNone(bool)
        self.assertIsInstance(bool, Expression)
        self.assertEqual('true', bool.evaluate())

        string = ExpressionType.string('toto')
        self.assertIsNotNone(string)
        self.assertIsInstance(string, Expression)
        self.assertEqual("'toto'", string.evaluate())

        integer = ExpressionType.number(101)
        self.assertIsNotNone(integer)
        self.assertIsInstance(integer, Expression)
        self.assertEqual('101', integer.evaluate())

        array = ExpressionType.collection([1, 2, 3, 4, 5])
        self.assertIsNotNone(array)
        self.assertIsInstance(array, Expression)
        self.assertEqual('[1,2,3,4,5]', array.evaluate())

        property = ExpressionType.property('Odata.CSC.my_prop')
        self.assertIsNotNone(property)
        self.assertIsInstance(property, Expression)
        self.assertEqual('Odata.CSC.my_prop', property.evaluate())

    def test_comparaison_operator(self):
        equal = ComparisonOperator.eq(
            ExpressionType.property('My_Prop'),
            ExpressionType.string('Toto')
        )
        self.assertIsNotNone(equal)
        self.assertIsInstance(equal, Expression)
        self.assertEqual("My_Prop eq 'Toto'", equal.evaluate())

        not_equal = ComparisonOperator.ne(
            ExpressionType.property('My_Prop'),
            ExpressionType.number(100)
        )
        self.assertIsNotNone(not_equal)
        self.assertIsInstance(not_equal, Expression)
        self.assertEqual("My_Prop ne 100", not_equal.evaluate())

        has = ComparisonOperator.has(
            ExpressionType.property('Style'),
            ExpressionType.property("Sales.Color'Yellow'")
        )
        self.assertIsNotNone(has)
        self.assertIsInstance(has, Expression)
        self.assertEqual("Style has Sales.Color'Yellow'", has.evaluate())

        is_member = ComparisonOperator.co_in(
            ExpressionType.property('Address/City'),
            GroupingOperator.group(
                ExpressionType.property("'Redmond', 'London'"))
        )
        self.assertIsNotNone(is_member)
        self.assertIsInstance(is_member, Expression)
        self.assertEqual("Address/City in ('Redmond', 'London')",
                         is_member.evaluate())

        lt = ComparisonOperator.lt(
            ExpressionType.property('Price'),
            ExpressionType.number(20)
        )
        self.assertIsNotNone(lt)
        self.assertIsInstance(lt, Expression)
        self.assertEqual("Price lt 20", lt.evaluate())

        le = ComparisonOperator.le(
            ExpressionType.property('Price'),
            ExpressionType.number(100)
        )
        self.assertIsNotNone(le)
        self.assertIsInstance(le, Expression)
        self.assertEqual("Price le 100", le.evaluate())

        gt = ComparisonOperator.gt(
            ExpressionType.property('Price'),
            ExpressionType.number(20)
        )
        self.assertIsNotNone(gt)
        self.assertIsInstance(gt, Expression)
        self.assertEqual("Price gt 20", gt.evaluate())

        ge = ComparisonOperator.ge(
            ExpressionType.property('Price'),
            ExpressionType.number(100)
        )
        self.assertIsNotNone(ge)
        self.assertIsInstance(ge, Expression)
        self.assertEqual("Price ge 100", ge.evaluate())

    def test_lo_operator(self):
        ge = ComparisonOperator.ge(
            ExpressionType.property('Price'),
            ExpressionType.number(100)
        )
        gt = ComparisonOperator.gt(
            ExpressionType.property('Price'),
            ExpressionType.number(20)
        )

        ope = LogicalOperator.lo_and(ge, gt)
        self.assertIsNotNone(ope)
        self.assertIsInstance(ope, Expression)
        self.assertEqual("Price ge 100 and Price gt 20", ope.evaluate())

        ope = LogicalOperator.lo_or(ge, gt)
        self.assertIsNotNone(ope)
        self.assertIsInstance(ope, Expression)
        self.assertEqual("Price ge 100 or Price gt 20", ope.evaluate())

        ope = LogicalOperator.lo_not(ge)
        self.assertIsNotNone(ope)
        self.assertIsInstance(ope, Expression)
        self.assertEqual("not Price ge 100", ope.evaluate())

    def test_arithmetics_operator(self):
        add = ArithmeticOperator.add(
            ExpressionType.property('Price'),
            ExpressionType.number(5))
        self.assertIsNotNone(add)
        self.assertIsInstance(add, Expression)
        self.assertEqual("Price add 5", add.evaluate())

        sub = ArithmeticOperator.sub(
            ExpressionType.property('Price'),
            ExpressionType.number(5))
        self.assertIsNotNone(sub)
        self.assertIsInstance(sub, Expression)
        self.assertEqual("Price sub 5", sub.evaluate())

        divby = ArithmeticOperator.divby(
            ExpressionType.property('Price'),
            ExpressionType.number(5))
        self.assertIsNotNone(divby)
        self.assertIsInstance(divby, Expression)
        self.assertEqual("Price divby 5", divby.evaluate())

        mul = ArithmeticOperator.mul(
            ExpressionType.property('Price'),
            ExpressionType.number(5))
        self.assertIsNotNone(mul)
        self.assertIsInstance(mul, Expression)
        self.assertEqual("Price mul 5", mul.evaluate())

        div = ArithmeticOperator.div(
            ExpressionType.property('Price'),
            ExpressionType.number(5))
        self.assertIsNotNone(div)
        self.assertIsInstance(div, Expression)
        self.assertEqual("Price div 5", div.evaluate())

        mod = ArithmeticOperator.mod(
            ExpressionType.property('Price'),
            ExpressionType.number(5))
        self.assertIsNotNone(mod)
        self.assertIsInstance(mod, Expression)
        self.assertEqual("Price mod 5", mod.evaluate())

    def test_grouping(self):
        group = ComparisonOperator.gt(
            GroupingOperator.group(
                ArithmeticOperator.add(
                    ExpressionType.property('Price'),
                    ExpressionType.number(5))
                ),
            ExpressionType.number(50)
            )
        self.assertIsNotNone(group)
        self.assertIsInstance(group, Expression)
        self.assertEqual("(Price add 5) gt 50", group.evaluate())

    def test_string_collection_function(self):
        concat = ExpressionFunc.concat(
            ExpressionType.property('City'),
            ExpressionType.string(', ')
        )
        self.assertIsNotNone(concat)
        self.assertIsInstance(concat, Expression)
        self.assertEqual("concat(City,', ')", concat.evaluate())

        contains = ExpressionFunc.contains(
            ExpressionType.property('City'),
            'London'
        )
        self.assertIsNotNone(contains)
        self.assertIsInstance(contains, Expression)
        self.assertEqual("contains(City,'London')", contains.evaluate())

        start = ExpressionFunc.startswith(
            ExpressionType.property('City'),
            'Lo'
        )
        self.assertIsNotNone(start)
        self.assertIsInstance(start, Expression)
        self.assertEqual("startswith(City,'Lo')", start.evaluate())

        end = ExpressionFunc.endswith(
            ExpressionType.property('City'),
            'on'
        )
        self.assertIsNotNone(end)
        self.assertIsInstance(end, Expression)
        self.assertEqual("endswith(City,'on')", end.evaluate())

        index = ExpressionFunc.indexof(
            ExpressionType.property('Company'),
            'Gael'
        )
        self.assertIsNotNone(index)
        self.assertIsInstance(index, Expression)
        self.assertEqual("indexof(Company,'Gael')", index.evaluate())

        length = ExpressionFunc.length(
            ExpressionType.property('Company')
        )
        self.assertIsNotNone(length)
        self.assertIsInstance(length, Expression)
        self.assertEqual("length(Company)", length.evaluate())

        substring = ExpressionFunc.substring(
            ExpressionType.property('Company'),
            1
        )
        eq = ComparisonOperator.eq(
            substring,
            ExpressionType.string('Gael Systems')
        )
        self.assertIsNotNone(eq)
        self.assertIsInstance(eq, Expression)
        self.assertEqual("substring(Company,1) eq 'Gael Systems'",
                         eq.evaluate())

    def test_collection_func(self):
        hasssubset = ExpressionFunc.hassubset(
            ExpressionType.collection([1, 2, 3, 4, 5]),
            ExpressionType.collection([3, 4])
        )
        self.assertIsNotNone(hasssubset)
        self.assertIsInstance(hasssubset, Expression)
        self.assertEqual("hassubset([1,2,3,4,5],[3,4])", hasssubset.evaluate())

        hassubsequence = ExpressionFunc.hassubsequence(
            ExpressionType.collection([1, 2, 3, 4, 5]),
            ExpressionType.collection([3, 5])
        )
        self.assertIsNotNone(hassubsequence)
        self.assertIsInstance(hassubsequence, Expression)
        self.assertEqual("hassubsequence([1,2,3,4,5],[3,5])",
                         hassubsequence.evaluate())

    def test_string_function(self):
        match = ExpressionFunc.matchesPattern(
            ExpressionType.property('CompanyName'),
            ExpressionType.string('%5EA.*e$')
        )
        self.assertIsNotNone(match)
        self.assertIsInstance(match, Expression)
        self.assertEqual("matchesPattern(CompanyName,'%5EA.*e$')",
                         match.evaluate())

        lower = ExpressionFunc.tolower(
            ExpressionType.property('CompanyName')
        )
        self.assertIsNotNone(lower)
        self.assertIsInstance(lower, Expression)
        self.assertEqual("tolower(CompanyName)", lower.evaluate())

        upper = ExpressionFunc.toupper(
            ExpressionType.property('CompanyName')
        )
        self.assertIsNotNone(upper)
        self.assertIsInstance(upper, Expression)
        self.assertEqual("toupper(CompanyName)", upper.evaluate())

        trim = ExpressionFunc.trim(
            ExpressionType.property('CompanyName')
        )
        self.assertIsNotNone(trim)
        self.assertIsInstance(trim, Expression)
        self.assertEqual("trim(CompanyName)", trim.evaluate())

        eq = ComparisonOperator.eq(
            upper,
            ExpressionType.string('GAEL SYSTEMSS')
        )
        self.assertIsNotNone(eq)
        self.assertIsInstance(eq, Expression)
        self.assertEqual("toupper(CompanyName) eq 'GAEL SYSTEMSS'",
                         eq.evaluate())

    def test_date_time_function(self):
        day = ExpressionFunc.day(
            ExpressionType.property('StartTime')
        )
        self.assertIsNotNone(day)
        self.assertIsInstance(day, Expression)
        self.assertEqual("day(StartTime)", day.evaluate())

        date = ExpressionFunc.date(
            ExpressionType.property('StartTime')
        )
        self.assertIsNotNone(date)
        self.assertIsInstance(date, Expression)
        self.assertEqual("date(StartTime)", date.evaluate())

        second = ExpressionFunc.second(
            ExpressionType.property('StartTime')
        )
        self.assertIsNotNone(second)
        self.assertIsInstance(second, Expression)
        self.assertEqual("second(StartTime)", second.evaluate())

        hour = ExpressionFunc.hour(
            ExpressionType.property('StartTime')
        )
        self.assertIsNotNone(hour)
        self.assertIsInstance(hour, Expression)
        self.assertEqual("hour(StartTime)", hour.evaluate())

        minute = ExpressionFunc.minute(
            ExpressionType.property('StartTime')
        )
        self.assertIsNotNone(minute)
        self.assertIsInstance(minute, Expression)
        self.assertEqual("minute(StartTime)", minute.evaluate())

        month = ExpressionFunc.month(
            ExpressionType.property('StartTime')
        )
        self.assertIsNotNone(month)
        self.assertIsInstance(month, Expression)
        self.assertEqual("month(StartTime)", month.evaluate())

        time = ExpressionFunc.time(
            ExpressionType.property('StartTime')
        )
        self.assertIsNotNone(time)
        self.assertIsInstance(time, Expression)
        self.assertEqual("time(StartTime)", time.evaluate())

        totaloffsetminutes = ExpressionFunc.totaloffsetminutes(
            ExpressionType.property('StartTime')
        )
        self.assertIsNotNone(totaloffsetminutes)
        self.assertIsInstance(totaloffsetminutes, Expression)
        self.assertEqual("totaloffsetminutes(StartTime)",
                         totaloffsetminutes.evaluate())

        year = ExpressionFunc.year(
            ExpressionType.property('StartTime')
        )
        self.assertIsNotNone(year)
        self.assertIsInstance(year, Expression)
        self.assertEqual("year(StartTime)", year.evaluate())

        maxdatetime = ExpressionFunc.maxdatetime(
            ExpressionType.property('StartTime')
        )
        self.assertIsNotNone(maxdatetime)
        self.assertIsInstance(maxdatetime, Expression)
        self.assertEqual("maxdatetime(StartTime)", maxdatetime.evaluate())

        mindatetime = ExpressionFunc.mindatetime(
            ExpressionType.property('StartTime')
        )
        self.assertIsNotNone(mindatetime)
        self.assertIsInstance(mindatetime, Expression)
        self.assertEqual("mindatetime(StartTime)", mindatetime.evaluate())

        totalseconds = ExpressionFunc.totalseconds(
            ExpressionType.property('StartTime')
        )
        self.assertIsNotNone(totalseconds)
        self.assertIsInstance(totalseconds, Expression)
        self.assertEqual("totalseconds(StartTime)", totalseconds.evaluate())

        now = ExpressionFunc.now()
        self.assertIsNotNone(now)
        self.assertIsInstance(now, Expression)
        self.assertEqual("now()", now.evaluate())

    def test_arithmetic_function(self):
        ceilling = ExpressionFunc.ceiling(
            ExpressionType.property('Freight')
        )
        self.assertIsNotNone(ceilling)
        self.assertIsInstance(ceilling, Expression)
        self.assertEqual("ceiling(Freight)", ceilling.evaluate())

        floor = ExpressionFunc.floor(
            ExpressionType.property('Freight')
        )
        self.assertIsNotNone(floor)
        self.assertIsInstance(floor, Expression)
        self.assertEqual("floor(Freight)", floor.evaluate())

        round = ExpressionFunc.round(
            ExpressionType.property('Freight')
        )
        self.assertIsNotNone(round)
        self.assertIsInstance(round, Expression)
        self.assertEqual("round(Freight)", round.evaluate())

        eq = ComparisonOperator.eq(
            round,
            ExpressionType.number(32)
        )
        self.assertIsNotNone(eq)
        self.assertIsInstance(eq, Expression)
        self.assertEqual("round(Freight) eq 32", eq.evaluate())

    def test_type_function(self):
        cast = ExpressionFunc.cast(
            ExpressionType.property('ShipCountry'),
            ExpressionType.property('Edm.String')
        )
        self.assertIsNotNone(cast)
        self.assertIsInstance(cast, Expression)
        self.assertEqual("cast(ShipCountry,Edm.String)", cast.evaluate())

        isof = ExpressionFunc.isof(
            ExpressionType.property('ShipCountry')
        )
        self.assertIsNotNone(isof)
        self.assertIsInstance(isof, Expression)
        self.assertEqual("isof(ShipCountry)", isof.evaluate())

        isof = ExpressionFunc.isof(
            ExpressionType.property('ShipCountry'),
            ExpressionType.property('Edm.String')
        )
        self.assertIsNotNone(isof)
        self.assertIsInstance(isof, Expression)
        self.assertEqual("isof(ShipCountry,Edm.String)", isof.evaluate())

    def test_geo_function(self):
        footprint = ExpressionType.footprint(self.foot_print)
        self.assertIsNotNone(footprint)
        self.assertIsInstance(footprint, Expression)
        self.assertEqual(self.foot_print,
                         footprint.evaluate())

        footprint = ExpressionType.footprint(
            Polygon([[0, 0], [1, 0], [1, 1], [0, 1], [0, 0]])
        )
        self.assertIsNotNone(footprint)
        self.assertIsInstance(footprint, Expression)
        self.assertEqual("geography'SRID=4326;"
                         "Polygon((0.0 0.0,1.0 0.0,1.0 1.0,0.0 1.0,0.0 0.0))'",
                         footprint.evaluate())

        footprint = ExpressionType.footprint(
            [(-12, 34), (32, 34), (32, 71), (-12, 71), (-12, 34)])
        self.assertIsNotNone(footprint)
        self.assertIsInstance(footprint, Expression)
        self.assertEqual("geography'SRID=4326;Polygon((-12 34,32 34,32 "
                         "71,-12 71,-12 34))'",
                         footprint.evaluate())

        distance = ExpressionFunc.geo_distance(
            ExpressionType.property('CurrentPosition'),
            footprint
        )
        self.assertIsNotNone(distance)
        self.assertIsInstance(distance, Expression)
        self.assertEqual("geo.distance(CurrentPosition,geography'SRID=4326;"
                         "Polygon((-12 34,32 34,32 71,-12 71,-12 34))')",
                         distance.evaluate())

        intersects = ExpressionFunc.geo_intersects(
            ExpressionType.property('CurrentPosition'),
            ExpressionType.property('TargetArea')
        )
        self.assertIsNotNone(intersects)
        self.assertIsInstance(intersects, Expression)
        self.assertEqual("geo.intersects(CurrentPosition,TargetArea)",
                         intersects.evaluate())

        length = ExpressionFunc.geo_length(
            ExpressionType.property('DirectRoute')
        )
        self.assertIsNotNone(length)
        self.assertIsInstance(length, Expression)
        self.assertEqual("geo.length(DirectRoute)", length.evaluate())

    def test_case_function(self):
        values = [
            (
                ComparisonOperator.gt(
                    ExpressionType.property('X'),
                    ExpressionType.number(0)
                ),
                1),
            (ComparisonOperator.lt(
                ExpressionType.property('X'),
                ExpressionType.number(0)
            ),
             -1),
            (
                ExpressionType.bool(True),
                0
            )
        ]
        case = ExpressionFunc.case(values)
        self.assertIsNotNone(case)
        self.assertIsInstance(case, Expression)
        self.assertEqual("case(X gt 0:1,X lt 0:-1,true:0)", case.evaluate())

    def test_lambda_function(self):
        any = ExpressionFunc.any(
            'd',
            ComparisonOperator.gt(
                ExpressionType.property('d/TotalPrice'),
                ExpressionType.number(100)
            )
        )
        self.assertIsNotNone(any)
        self.assertIsInstance(any, Expression)
        self.assertEqual("any(d:d/TotalPrice gt 100)", any.evaluate())

        all = ExpressionFunc.all(
            'a',
            ComparisonOperator.gt(
                ExpressionType.property('a/TotalPrice'),
                ExpressionType.number(100)
            )
        )
        self.assertIsNotNone(all)
        self.assertIsInstance(all, Expression)
        self.assertEqual("all(a:a/TotalPrice gt 100)", all.evaluate())



import os
import unittest
import httpretty

from drb.drivers.odata import (
    ODataServiceNodeCSC, ODataQueryPredicate,
    ExpressionFunc, ExpressionType)


def odata_metadata_query(request, uri, headers):
    stub = '<odata><foo><bar Namespace="OData.CSC"></bar></foo></odata>'
    return 200, headers, stub


class TestODataPredicate(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        os.path.join(os.path.dirname(__file__), 'resources')
        httpretty.enable()

    @classmethod
    def tearDownClass(cls) -> None:
        httpretty.disable()
        httpretty.reset()

    def test_custom_query(self):
        url = 'http://gael-systems.com/odata'
        my_filter = "startswith(Name,'S2')"
        my_order = "ContentLength desc"
        products_url = f'{url}/Products'

        def callback_checker(request, uri, headers):
            self.assertIn('$filter', request.querystring.keys())
            self.assertIn(my_filter, request.querystring['$filter'][0])
            self.assertIn('$orderby', request.querystring.keys())
            self.assertIn(request.querystring['$orderby'][0],
                          [
                              my_order,
                              f"{my_order} PublicationDate asc"
                          ])
            self.assertIn('$format', request.querystring.keys())
            self.assertEqual('json', request.querystring['$format'][0])
            if '$count' in request.querystring \
                    and request.querystring['$count'][0]:
                return 200, headers, '{"@odata.count": 0,"value": []}'
            return 200, headers, '{"value": []}'

        httpretty.register_uri(httpretty.GET, f'{url}/$metadata',
                               odata_metadata_query)
        httpretty.register_uri(httpretty.GET, products_url, callback_checker)

        node = ODataServiceNodeCSC(url)
        children = node[
            ODataQueryPredicate(filter=my_filter, order=my_order)]

        self.assertIsNotNone(children)
        self.assertIsInstance(children, list)
        self.assertEqual(0, len(children))

        children = node[
            ODataQueryPredicate(
                filter=ExpressionFunc.startswith(
                                    ExpressionType.property('Name'),
                                    'S2'
                                ),
                order=(
                    ExpressionType.property('ContentLength'),
                    ExpressionType.property('desc')
                )
            )]

        self.assertIsNotNone(children)
        self.assertIsInstance(children, list)
        self.assertEqual(0, len(children))

        children = node[
            ODataQueryPredicate(
                filter=ExpressionFunc.startswith(
                                    ExpressionType.property('Name'),
                                    'S2'
                                ).evaluate(),
                order=[
                    (
                        ExpressionType.property('ContentLength'),
                        ExpressionType.property('desc')
                    ),
                    (
                        ExpressionType.property('PublicationDate'),
                        ExpressionType.property('asc')
                    )
                    ]
            )]

        self.assertIsNotNone(children)
        self.assertIsInstance(children, list)
        self.assertEqual(0, len(children))

        children = node[ODataQueryPredicate(filter=my_filter, order=my_order)]
        self.assertIsNotNone(children)
        self.assertIsInstance(children, list)
        self.assertEqual(0, len(children))

    def test_custom_query_with_count(self):
        url = 'http://gael-systems.com/odata2'
        my_filter = "startswith(Name,'S2')"
        my_order = "ContentLength desc"
        products_url = f'{url}/Products'

        def callback_checker(request, uri, headers):
            self.assertIn('$filter', request.querystring.keys())
            self.assertIn(my_filter, request.querystring['$filter'][0])
            self.assertIn('$orderby', request.querystring.keys())
            self.assertIn(my_order, request.querystring['$orderby'][0])
            self.assertIn('$format', request.querystring.keys())
            self.assertEqual('json', request.querystring['$format'][0])
            self.assertIn('$count', request.querystring.keys())
            self.assertTrue('true', request.querystring['$count'][0])
            return 200, headers, '{"@odata.count": 100, "value": []}'

        httpretty.register_uri(httpretty.GET, f'{url}/$metadata',
                               odata_metadata_query)
        httpretty.register_uri(httpretty.GET, products_url, callback_checker)

        node = ODataServiceNodeCSC(url)
        children = node[ODataQueryPredicate(filter=my_filter, order=my_order)]

        self.assertIsNotNone(children)
        self.assertIsInstance(children, list)
        self.assertEqual(100, len(children))

        children = node[ODataQueryPredicate(filter=my_filter, order=my_order)]
        self.assertIsNotNone(children)
        self.assertIsInstance(children, list)
        self.assertEqual(100, len(children))

    def test_custom_query_without_count(self):
        url = 'http://gael-systems.com/odata'
        my_filter = "startswith(Name,'S2')"
        my_order = "ContentLength desc"
        products_url = f'{url}/Products'

        def callback_checker(request, uri, headers):
            self.assertIn('$filter', request.querystring.keys())
            self.assertIn(my_filter, request.querystring['$filter'][0])
            self.assertIn('$orderby', request.querystring.keys())
            self.assertEqual(my_order, request.querystring['$orderby'][0])
            self.assertIn('$format', request.querystring.keys())
            self.assertEqual('json', request.querystring['$format'][0])
            self.assertIn('$count', request.querystring.keys())
            return 200, headers, '{"value": [], "@odata.count": 0}'

        httpretty.register_uri(httpretty.GET, f'{url}/$metadata',
                               odata_metadata_query)
        httpretty.register_uri(httpretty.GET, products_url, callback_checker)

        node = ODataServiceNodeCSC(url)
        children = node[ODataQueryPredicate(filter=my_filter, order=my_order)]

        self.assertIsNotNone(children)
        self.assertIsInstance(children, list)
        self.assertEqual(0, len(children))

        children = node[ODataQueryPredicate(filter=my_filter, order=my_order)]
        self.assertIsNotNone(children)
        self.assertIsInstance(children, list)
        self.assertEqual(0, len(children))



import abc
import re
from typing import Union

from shapely import Polygon


class Expression(abc.ABC):
    @abc.abstractmethod
    def evaluate(self) -> str:
        raise NotImplementedError


class BinaryExpr(Expression):
    def __init__(self, name: str, left: Expression, right: Expression):
        super().__init__()
        self._name = name
        self._left = left
        self._right = right

    def evaluate(self) -> str:
        return f"{self._left.evaluate()} {self._name} {self._right.evaluate()}"


class FunctionExpr(Expression):
    def __init__(self, name, *args):
        super().__init__()
        self._name = name
        self._args = [expr.evaluate() for expr in args]

    def evaluate(self) -> str:
        return f"{self._name}({','.join(self._args)})"


class NotExpr(Expression):
    def __init__(self, expr: Expression):
        super().__init__()
        self._expr = expr

    def evaluate(self) -> str:
        return f"not {self._expr.evaluate()}"


class GroupExpr(Expression):
    def __init__(self, expr):
        super().__init__()
        self._expr = expr

    def evaluate(self) -> str:
        return f"({self._expr.evaluate()})"


class PropExpr(Expression):
    def __init__(self, name):
        super().__init__()
        self.name = name

    def evaluate(self) -> str:
        return self.name


class BooleanExpr(Expression):
    def __init__(self, value: bool):
        super().__init__()
        self.value = value

    def evaluate(self) -> str:
        return 'true' if self.value else 'false'


class NumberExpr(Expression):
    def __init__(self, value: Union[int, float]):
        super().__init__()
        self.value = value

    def evaluate(self) -> str:
        return str(self.value)


class StringExpr(Expression):
    def __init__(self, value):
        super().__init__()
        self.value = value

    def evaluate(self) -> str:
        return f"'{self.value}'"


class CollectionExpr(Expression):
    def __init__(self, value: list):
        super().__init__()
        self.value = value

    def evaluate(self) -> str:
        return f"{self.value}".replace(" ", "")


class Footprint(Expression):
    __pattern = re.compile(r"geography'SRID=4326;Polygon\(\((.+)\)\)'")

    @classmethod
    def from_data(cls, data) -> list:
        p = cls.__pattern.match(data)
        points = []
        if p is not None:
            for coords in p.group(1).split(','):
                lon, lat = coords.split(' ')
                points.append((lon, lat))
        return points

    def __init__(self, points: Union[list, str, Polygon]):
        if isinstance(points, str):
            self._points = self.from_data(points)
        elif isinstance(points, Polygon):
            self._points = [point for point in points.exterior.coords]
        else:
            self._points = points

    def evaluate(self) -> str:
        return "geography'SRID=4326;Polygon(" \
               f"({','.join([f'{p[0]} {p[1]}' for p in self._points])}))'"


class ExpressionType:
    @staticmethod
    def bool(value: bool) -> Expression:
        return BooleanExpr(value)

    @staticmethod
    def string(value: str) -> Expression:
        return StringExpr(value)

    @staticmethod
    def number(value: Union[int, float]):
        return NumberExpr(value)

    @staticmethod
    def collection(value: list):
        return CollectionExpr(value)

    @staticmethod
    def property(value: str):
        return PropExpr(value)

    @staticmethod
    def footprint(points: Union[list, str]):
        return Footprint(points)

    @staticmethod
    def _return(exp: Expression, value):
        if not isinstance(value, str):
            value = str(value)
        return PropExpr(exp.evaluate()+':'+value)


class ComparisonOperator:
    @staticmethod
    def eq(left: Expression, right: Expression) -> Expression:
        return BinaryExpr('eq', left, right)

    @staticmethod
    def ne(left: Expression, right: Expression) -> Expression:
        return BinaryExpr('ne', left, right)

    @staticmethod
    def has(left: Expression, right: Expression) -> Expression:
        return BinaryExpr('has', left, right)

    @staticmethod
    def co_in(left: Expression, right: Expression) -> Expression:
        return BinaryExpr('in', left, right)

    @staticmethod
    def lt(left: Expression, right: Expression) -> Expression:
        return BinaryExpr('lt', left, right)

    @staticmethod
    def le(left: Expression, right: Expression) -> Expression:
        return BinaryExpr('le', left, right)

    @staticmethod
    def gt(left: Expression, right: Expression) -> Expression:
        return BinaryExpr('gt', left, right)

    @staticmethod
    def ge(left: Expression, right: Expression) -> Expression:
        return BinaryExpr('ge', left, right)


class LogicalOperator:
    # Logical Operators
    @staticmethod
    def lo_and(left: Expression, right: Expression) -> Expression:
        return BinaryExpr('and', left, right)

    @staticmethod
    def lo_or(left: Expression, right: Expression) -> Expression:
        return BinaryExpr('or', left, right)

    @staticmethod
    def lo_not(expr: Expression) -> Expression:
        return NotExpr(expr)


class ArithmeticOperator:
    # Arithmetic Operators
    @staticmethod
    def add(left: Expression, right: Expression) -> Expression:
        return BinaryExpr('add', left, right)

    @staticmethod
    def sub(left: Expression, right: Expression) -> Expression:
        return BinaryExpr('sub', left, right)

    @staticmethod
    def mul(left: Expression, right: Expression) -> Expression:
        return BinaryExpr('mul', left, right)

    @staticmethod
    def div(left: Expression, right: Expression) -> Expression:
        return BinaryExpr('div', left, right)

    @staticmethod
    def divby(left: Expression, right: Expression) -> Expression:
        return BinaryExpr('divby', left, right)

    @staticmethod
    def mod(left: Expression, right: Expression) -> Expression:
        return BinaryExpr('mod', left, right)


class GroupingOperator:
    @staticmethod
    def group(value: Expression):
        return GroupExpr(value)


class ExpressionFunc:
    # String and Collection Functions
    @staticmethod
    def concat(expr: Expression, expr_2: Expression) -> Expression:
        return FunctionExpr('concat', expr, expr_2)

    @staticmethod
    def contains(expr: Expression, content: str) -> Expression:
        return FunctionExpr('contains', expr,
                            ExpressionType.string(content))

    @staticmethod
    def endswith(expr: Expression, suffix: str) -> Expression:
        return FunctionExpr('endswith', expr, ExpressionType.string(suffix))

    @staticmethod
    def startswith(expr: Expression, prefix: str) -> Expression:
        return FunctionExpr('startswith', expr,
                            ExpressionType.string(prefix))

    @staticmethod
    def indexof(expr: Expression, value: str) -> Expression:
        return FunctionExpr('indexof', expr, ExpressionType.string(value))

    @staticmethod
    def length(expr: Expression) -> Expression:
        return FunctionExpr('length', expr)

    @staticmethod
    def substring(expr: Expression, index: Union[int, float]) -> Expression:
        return FunctionExpr('substring', expr, NumberExpr(index))

    # Collection Functions
    @staticmethod
    def hassubset(expr: list, expr_2: list,) -> Expression:
        return FunctionExpr('hassubset',
                            expr,
                            expr_2)

    @staticmethod
    def hassubsequence(expr: Expression, expr_2: Expression,) -> Expression:
        return FunctionExpr('hassubsequence',
                            expr,
                            expr_2
                            )

    # String Functions

    @staticmethod
    def matchesPattern(expr: Expression, expr_2: Expression) -> Expression:
        return FunctionExpr('matchesPattern', expr, expr_2)

    @staticmethod
    def tolower(expr: Expression) -> Expression:
        return FunctionExpr('tolower', expr)

    @staticmethod
    def toupper(expr: Expression) -> Expression:
        return FunctionExpr('toupper', expr)

    @staticmethod
    def trim(expr: Expression) -> Expression:
        return FunctionExpr('trim', expr)

    # Date and Time Functions

    @staticmethod
    def day(expr: Expression) -> Expression:
        return FunctionExpr('day', expr)

    @staticmethod
    def date(expr: Expression) -> Expression:
        return FunctionExpr('date', expr)

    @staticmethod
    def second(expr: Expression) -> Expression:
        return FunctionExpr('second', expr)

    @staticmethod
    def hour(expr: Expression) -> Expression:
        return FunctionExpr('hour', expr)

    @staticmethod
    def minute(expr: Expression) -> Expression:
        return FunctionExpr('minute', expr)

    @staticmethod
    def month(expr: Expression) -> Expression:
        return FunctionExpr('month', expr)

    @staticmethod
    def time(expr: Expression) -> Expression:
        return FunctionExpr('time', expr)

    @staticmethod
    def totaloffsetminutes(expr: Expression) -> Expression:
        return FunctionExpr('totaloffsetminutes', expr)

    @staticmethod
    def totalseconds(expr: Expression) -> Expression:
        return FunctionExpr('totalseconds', expr)

    @staticmethod
    def year(expr: Expression) -> Expression:
        return FunctionExpr('year', expr)

    @staticmethod
    def maxdatetime(expr: Expression) -> Expression:
        return FunctionExpr('maxdatetime', expr)

    @staticmethod
    def mindatetime(expr: Expression) -> Expression:
        return FunctionExpr('mindatetime', expr)

    @staticmethod
    def now() -> Expression:
        return FunctionExpr('now')

    # Arithmetic Functions
    @staticmethod
    def ceiling(expr: Expression) -> Expression:
        return FunctionExpr('ceiling', expr)

    @staticmethod
    def floor(expr: Expression) -> Expression:
        return FunctionExpr('floor', expr)

    @staticmethod
    def round(expr: Expression) -> Expression:
        return FunctionExpr('round', expr)

    # Type Functions

    @staticmethod
    def cast(expr: Expression, expr_2: Expression) -> Expression:
        return FunctionExpr('cast', expr, expr_2)

    @staticmethod
    def isof(expr: Expression, expr_2: Expression = None) -> Expression:
        if expr_2 is None:
            return FunctionExpr('isof', expr)
        return FunctionExpr('isof', expr, expr_2)

    # Geo Functions
    @staticmethod
    def geo_distance(expr: Expression, expr_2: Expression = None)\
            -> Expression:
        return FunctionExpr('geo.distance', expr, expr_2)

    @staticmethod
    def geo_intersects(expr: Expression, expr_2: Expression = None)\
            -> Expression:
        return FunctionExpr('geo.intersects', expr, expr_2)

    @staticmethod
    def geo_length(expr: Expression) -> Expression:
        return FunctionExpr('geo.length', expr)

    @staticmethod
    def csc_intersect(expr: Expression, expr_2: Expression):
        return FunctionExpr('OData.CSC.Intersects', expr, expr_2)

    # Conditional Functions
    @staticmethod
    def case(values: list):
        tmp = tuple([ExpressionType._return(e[0], e[1]) for e in values])
        str = ''
        for t in tmp:
            str += t.evaluate()+','
        return FunctionExpr('case',
                            ExpressionType.property(str[:len(str)-1]))

    # lambda
    @staticmethod
    def any(var: str, exp: Expression):
        return FunctionExpr('any', PropExpr(f'{var}:{exp.evaluate()}'))

    @staticmethod
    def all(var: str, exp: Expression):
        return FunctionExpr('all', PropExpr(f'{var}:{exp.evaluate()}'))



import uuid
import unittest

from drb.drivers.odata import ODataServiceNodeCSC, \
    ODataProductNode, ODataQueryPredicate
from tests.utils import start_mock_odata_csc, stop_mock_odata_csc


class TestODataServiceNode(unittest.TestCase):
    svc_url = 'https://gael-systems.com/odata/csc'
    node = None

    @classmethod
    def setUpClass(cls) -> None:
        start_mock_odata_csc(cls.svc_url)
        cls.node = ODataServiceNodeCSC(cls.svc_url)

    @classmethod
    def tearDownClass(cls) -> None:
        stop_mock_odata_csc()

    def test_index(self):
        child = self.node[0]
        self.assertIsNotNone(child)
        self.assertNotIsInstance(child, ODataProductNode)

    def test_slice(self):
        children = self.node[0:3]
        child = children[0]
        self.assertIsNotNone(children)
        self.assertIsInstance(children, list)
        self.assertIsNotNone(child)
        self.assertNotIsInstance(child, ODataProductNode)

    def test_uuid(self):
        prd_uuid = '0723d9bf-02a2-3e99-b1b3-f6d81de84b62'
        child = self.node[uuid.UUID(prd_uuid)]
        self.assertIsNotNone(child)
        self.assertNotIsInstance(child, ODataProductNode)

    def test_name(self):
        prd_name = "S2B_OPER_MSI_L0__GR_EPAE_20180703T214414_" \
                   "S20180703T165907_D05_N02.06.tar"
        child = self.node[prd_name]
        self.assertIsNotNone(child)
        self.assertNotIsInstance(child, ODataProductNode)

    def test_resolve(self):
        prd_name = "banana.png"
        child = self.node[prd_name]
        self.assertIsNotNone(child)
        self.assertIsInstance(child, ODataProductNode)

    def test_filter_predicate(self):
        prd_name = "S2B_OPER_MSI_L0__GR_EPAE_20180703T214414_" \
                   "S20180703T165907_D05_N02.06.tar"
        children = self.node[ODataQueryPredicate(
            filter=f"Name eq '{prd_name}'")]
        child = children[0]
        self.assertIsNotNone(child)
        self.assertNotIsInstance(child, ODataProductNode)

    def test_slice_children(self):
        children = self.node.children[0:3]
        self.assertIsNotNone(children)
        self.assertIsInstance(children, list)
        for i in range(2):
            child = children[i]
            self.assertIsNotNone(child)
            self.assertNotIsInstance(child, ODataProductNode)

        with self.assertRaises(IndexError):
            child = children[8]

        with self.assertRaises(KeyError):
            child = children['foo']

    def test_iter_children(self):
        children = self.node.children[0:3]
        itr = iter(children)
        for child in itr:
            self.assertIsNotNone(child)
            self.assertNotIsInstance(child, ODataProductNode)

    def test_append(self):
        children = self.node.children[0:3]
        with self.assertRaises(NotImplementedError):
            child = children.append('foo')

    def test_clear(self):
        children = self.node.children[0:3]
        with self.assertRaises(NotImplementedError):
            children.clear()

    def test_copy(self):
        children = self.node.children[0:3]
        with self.assertRaises(NotImplementedError):
            children.copy()

    def test_count(self) -> int:
        children = self.node.children[0:3]
        with self.assertRaises(NotImplementedError):
            children.count(10)

    def test_extend(self):
        children = self.node.children[0:3]
        with self.assertRaises(NotImplementedError):
            children.extend([])

    def test_insert(self):
        children = self.node.children[0:3]
        with self.assertRaises(NotImplementedError):
            children.insert(2, 'foo')

    def test_pop(self):
        children = self.node.children[0:3]
        with self.assertRaises(NotImplementedError):
            children.pop(2)

    def test_remove(self):
        children = self.node.children[0:3]
        with self.assertRaises(NotImplementedError):
            children.remove(10)

    def test_reverse(self):
        children = self.node.children[0:3]
        with self.assertRaises(NotImplementedError):
            children.reverse()

    def test_sort(self):
        children = self.node.children[0:3]
        with self.assertRaises(NotImplementedError):
            children.sort()



import json
import io
from enum import Enum

from drb.core import Predicate
from drb.drivers.http import DrbHttpNode
from drb.exceptions.core import DrbException
from requests.auth import AuthBase
from typing import List
from defusedxml import ElementTree
from defusedxml.ElementTree import ParseError

from .expression import Expression
from .odata_node import OdataNode
from ...exceptions.odata import OdataRequestException


class ODataServiceType(Enum):
    UNKNOWN = 0
    CSC = 1
    DHUS = 2
    ONDA_DIAS = 3


class ODataUtils:

    @staticmethod
    def http_node_to_json(node: DrbHttpNode) -> dict:
        try:
            with node.get_impl(io.BytesIO) as stream:
                data = json.load(stream)
                if 'error' in data.keys():
                    raise OdataRequestException(str(data['error']))
                return data
        except json.JSONDecodeError:
            raise OdataRequestException(f'Invalid json from {node.path.name}')
        except DrbException:
            raise OdataRequestException(f'Invalid node: {type(node)}')

    @staticmethod
    def get_type_odata_svc(service_url: str, auth: AuthBase = None) \
            -> ODataServiceType:
        """
        Retrieve with the given URL the OData service type (CSC or DHuS).

        Parameters:
            service_url (str): service URL
            auth (AuthBase): authentication mechanism required by the service
                             (default: ``None``)
        Returns:
            ODataServiceType: value corresponding to service
        """
        try:
            url = f'{service_url}/$metadata'
            node = DrbHttpNode(url, auth=auth)
            tree = ElementTree.parse(node.get_impl(io.BytesIO))
            ns = tree.getroot()[0][0].get('Namespace', None)
            if 'OData.CSC'.lower() == ns.lower():
                return ODataServiceType.CSC
            elif 'OData.DHuS'.lower() == ns.lower():
                return ODataServiceType.DHUS
            elif 'Ens'.lower() == ns.lower():
                return ODataServiceType.ONDA_DIAS
            return ODataServiceType.UNKNOWN
        except (DrbException, ParseError) as ex:
            return ODataServiceType.UNKNOWN

    @staticmethod
    def req_svc(odata: OdataNode) -> dict:
        node = DrbHttpNode(odata.get_service_url(), auth=odata.get_auth(),
                           params={'$format': 'json'})
        data = ODataUtils.http_node_to_json(node)
        return data

    @staticmethod
    def req_svc_count(odata: OdataNode) -> int:
        url = f'{odata.get_service_url()}/Products/$count'
        node = DrbHttpNode(url, auth=odata.get_auth())
        stream = node.get_impl(io.BytesIO)
        value = stream.read().decode()
        stream.close()
        return int(value)

    @staticmethod
    def req_svc_count_search(odata: OdataNode, search: str) -> int:
        url = f'{odata.get_service_url()}/Products/$count?$search={search}'
        node = DrbHttpNode(url, auth=odata.get_auth())
        stream = node.get_impl(io.BytesIO)
        value = stream.read().decode()
        stream.close()
        return int(value)

    @staticmethod
    def req_svc_products(odata: OdataNode, **kwargs) -> list:
        params = {'$format': 'json'}
        ret_count = False

        if 'filter' in kwargs.keys() and kwargs['filter'] is not None:
            params[odata.filter_keyword] = \
                kwargs['filter'].replace('\'', '%27')
        # For future use if we make search in GSS or Dhus...
        elif 'search' in kwargs.keys() and kwargs['search'] is not None:
            params[odata.filter_keyword] = kwargs['search']
        if 'order' in kwargs.keys() and kwargs['order'] is not None:
            params['$orderby'] = kwargs['order']

        if 'skip' in kwargs.keys() and kwargs['skip'] is not None:
            params['$skip'] = kwargs['skip']

        if 'top' in kwargs.keys() and kwargs['top'] is not None:
            params['$top'] = kwargs['top']

        if 'count' in kwargs.keys():
            ret_count = True
            count = kwargs['count']
            if count == -1 and odata.is_count_accepted_in_request:
                params['$count'] = 'true'

        query = '&'.join(map(lambda k: f'{k[0]}={k[1]}', params.items()))
        url = f'{odata.get_service_url()}/Products?{query}'
        node = DrbHttpNode(url, auth=odata.get_auth())
        data = ODataUtils.http_node_to_json(node)
        if ret_count:
            if '@odata.count' in data.keys():
                return data['value'], data['@odata.count']
            else:
                return data['value'], count
        return data['value']

    @staticmethod
    def req_svc_orders(odata: OdataNode, **kwargs) -> list:
        params = {'$format': 'json'}
        ret_count = False

        if 'filter' in kwargs.keys() and kwargs['filter'] is not None:
            params['filter'] = kwargs['filter'].replace('\'', '%27')
        # For future use if we make search in GSS or Dhus...
        if 'skip' in kwargs.keys() and kwargs['skip'] is not None:
            params['$skip'] = kwargs['skip']

        if 'top' in kwargs.keys() and kwargs['top'] is not None:
            params['$top'] = kwargs['top']

        if 'count' in kwargs.keys():
            ret_count = True
            count = kwargs['count']
            if count == -1:
                params['$count'] = 'true'

        query = '&'.join(map(lambda k: f'{k[0]}={k[1]}', params.items()))
        url = f'{odata.get_service_url()}/Orders?{query}'
        node = DrbHttpNode(url, auth=odata.get_auth())
        data = ODataUtils.http_node_to_json(node)
        if ret_count:
            if '@odata.count' in data.keys():
                return data['value'], data['@odata.count']
            else:
                return data['value'], count
        return data['value']

    @staticmethod
    def req_product_by_uuid(odata: OdataNode, prd_uuid: str) -> dict:
        url = f'{odata.get_service_url()}/{prd_uuid}'
        params = {'$format': 'json'}
        node = DrbHttpNode(url, auth=odata.get_auth(), params=params)
        return {
            k: v for k, v in ODataUtils.http_node_to_json(node).items()
            if not k.startswith('@odata.')
        }

    @staticmethod
    def req_product_order_by_uuid(odata):
        url = odata.path.original_path + '/Product'
        node = DrbHttpNode(url, auth=odata.parent.get_auth())
        return {
            k: v for k, v in ODataUtils.http_node_to_json(node).items()
            if not k.startswith('@odata.')
        }

    @staticmethod
    def req_product_attributes(odata: OdataNode,
                               prd_uuid: str,
                               attributes: str
                               ) -> List[dict]:
        url = f'{odata.get_service_url()}/{prd_uuid}/{attributes}'
        params = {'$format': 'json'}
        node = DrbHttpNode(url, auth=odata.get_auth(), params=params)
        data = ODataUtils.http_node_to_json(node)
        return data['value']

    @staticmethod
    def req_product_download(odata: OdataNode, prd_uuid: str, start=None,
                             end=None) -> io.BytesIO:
        url = f'{odata.get_service_url()}/{prd_uuid}/$value'
        node = DrbHttpNode(url, auth=odata.get_auth())
        if start is None or end is None:
            return node.get_impl(io.BytesIO)
        return node.get_impl(io.BytesIO, start=start, end=end)

    @staticmethod
    def launch_order(odata: OdataNode, prd_uuid: str, order: str):
        if not isinstance(prd_uuid, str):
            raise TypeError
        url = f'{odata.get_service_url()}/' \
              f'{odata.format_product(prd_uuid)}/{order}'
        node = DrbHttpNode.post(url,
                                auth=odata.get_auth())
        return node

    @staticmethod
    def get_order_by_uuid(odata: OdataNode, order_uuid: str):
        url = f'{odata.get_service_url()}/{order_uuid}'
        node = DrbHttpNode(url, auth=odata.get_auth())
        return ODataUtils.http_node_to_json(node)

    @staticmethod
    def cancel_order(odata: OdataNode, order_uuid: str):
        return DrbHttpNode.post(
            f"{odata.get_service_url()}/{order_uuid}/OData.CSC.Cancel",
            auth=odata.get_auth())


class ODataQueryPredicate(Predicate):
    """
    This predicate allowing to customize an OData query request.
    Customizable OData query elements: filter, search, orderby.

    Keyword Arguments:
        filter (str | Expression): the OData filter query element
        search (str): the OData search query element
        order (str): the OData orderby query element
    """

    def __init__(self, **kwargs):
        self.__filter = kwargs['filter'] if 'filter' in kwargs.keys() else None
        self.__search = kwargs['search'] if 'search' in kwargs.keys() else None
        self.__order = kwargs['order'] if 'order' in kwargs.keys() else None

    @property
    def filter(self) -> str:
        if isinstance(self.__filter, Expression):
            return self.__filter.evaluate()
        return self.__filter

    @property
    def order(self) -> str:
        if isinstance(self.__order, tuple):
            return f"{self.__order[0].evaluate()} {self.__order[1].evaluate()}"
        if isinstance(self.__order, list):
            order = ''
            for o in self.__order:
                order += f"{o[0].evaluate()} {o[1].evaluate()} "
            return order[:len(order)-1]
        return self.__order

    @property
    def search(self) -> str:
        return self.__search

    def matches(self, key) -> bool:
        return False



import io
import json
import os
import uuid
import unittest

from unittest.mock import patch

from drb.drivers.http import DrbHttpNode
from drb.nodes.logical_node import DrbLogicalNode

from drb.drivers.odata.odata_utils import ODataUtils, \
    ODataServiceType, ODataQueryPredicate
from drb.exceptions.odata import OdataRequestException


class TestODataUtils(unittest.TestCase):
    resource_dir = None

    @classmethod
    def setUpClass(cls) -> None:
        cls.resource_dir = os.path.join(os.path.dirname(__file__), 'resources')

    def test_get_type_odata_svc(self):
        svc_url = 'https://foobar.com/odata/v4'

        svc_metadata_path = os.path.join(self.resource_dir, 'csc_metadata.xml')
        with open(svc_metadata_path, 'rb') as resp:
            with patch.object(DrbHttpNode, 'get_impl', return_value=resp):
                t = ODataUtils.get_type_odata_svc(svc_url)
            self.assertEqual(ODataServiceType.CSC, t)

        svc_metadata_path = os.path.join(self.resource_dir,
                                         'dhus_metadata.xml')
        with open(svc_metadata_path, 'rb') as resp:
            with patch.object(DrbHttpNode, 'get_impl', return_value=resp):
                t = ODataUtils.get_type_odata_svc(svc_url)
            self.assertEqual(ODataServiceType.DHUS, t)

        svc_metadata_path = os.path.join(self.resource_dir,
                                         'dias_metadata.xml')
        with open(svc_metadata_path, 'rb') as resp:
            with patch.object(DrbHttpNode, 'get_impl', return_value=resp):
                t = ODataUtils.get_type_odata_svc(svc_url)
            self.assertEqual(ODataServiceType.ONDA_DIAS, t)

        svc_metadata_path = os.path.join(self.resource_dir,
                                         'no_csc_metadata.xml')
        with open(svc_metadata_path, 'rb') as resp:
            with patch.object(DrbHttpNode, 'get_impl', return_value=resp):
                t = ODataUtils.get_type_odata_svc(svc_url)
            self.assertEqual(ODataServiceType.UNKNOWN, t)

        with io.BytesIO(b'something else') as resp:
            with patch.object(DrbHttpNode, 'get_impl', return_value=resp):
                t = ODataUtils.get_type_odata_svc(svc_url)
            self.assertEqual(ODataServiceType.UNKNOWN, t)

    def test_http_node_to_json(self):
        url = 'https://my.domain.com'

        json_data = os.path.join(self.resource_dir, 'product1.json')
        with open(json_data, 'rb') as resp:
            with patch.object(DrbHttpNode, 'get_impl', return_value=resp):
                data = ODataUtils.http_node_to_json(DrbHttpNode(url))
        self.assertEqual(18388992, data['ContentLength'])
        self.assertEqual('0723d9a4-3bbe-305e-b712-5e820058e065', data['Id'])
        self.assertEqual(14, len(data))

        with io.BytesIO(b'<some><xml>content</xml><some>') as resp:
            with patch.object(DrbHttpNode, 'get_impl', return_value=resp):
                with self.assertRaises(OdataRequestException):
                    ODataUtils.http_node_to_json(DrbHttpNode(url))

        with self.assertRaises(OdataRequestException):
            ODataUtils.http_node_to_json(DrbLogicalNode('node'))

    def test_req_svc(self):
        expected = {
            "@odata.context": "$metadata",
            "value": [
                {"name": "Orders", "url": "Orders"},
                {"name": "Products", "url": "Products"}
            ]
        }
        with io.BytesIO(json.dumps(expected).encode()) as resp:
            with patch('drb.drivers.odata.odata_node.OdataNode') as odata:
                odata.get_service_url.return_value = 'https://test.com'
                odata.get_auth.return_value = None
                with patch.object(DrbHttpNode, 'get_impl', return_value=resp):
                    actual = ODataUtils.req_svc(odata)
        self.assertEqual(expected, actual)

    def test_req_svc_count(self):
        expected = 42
        with patch('drb.drivers.odata.odata_node.OdataNode') as odata:
            odata.get_service_url.return_value = 'https://test.com'
            odata.get_auth.return_value = None
            with io.BytesIO(str(expected).encode()) as resp:
                with patch.object(DrbHttpNode, 'get_impl', return_value=resp):
                    actual = ODataUtils.req_svc_count(odata)
        self.assertEqual(expected, actual)

    @patch('drb.drivers.odata.odata_node.OdataNode')
    def test_req_svc_products(self, odata):
        data = {
            '@odata.count': 132,
            'value': [
                {
                    'Id': str(uuid.uuid4()),
                    'Name': 'name#1',
                },
                {
                    'Id': str(uuid.uuid4()),
                    'Name': 'name#2',
                },
            ],
        }
        odata.get_service_url.return_value = 'https://my.domain.com/odata'
        odata.get_auth.return_value = None

        with io.BytesIO(json.dumps(data).encode()) as resp:
            with patch.object(DrbHttpNode, 'get_impl', return_value=resp):
                actual = ODataUtils.req_svc_products(odata,
                                                     filter='criteria#filter',
                                                     skip=1, top=2, order='Id')
        self.assertEqual(data['value'][0], actual[0])

        with io.BytesIO(json.dumps(data).encode()) as resp:
            with patch.object(DrbHttpNode, 'get_impl', return_value=resp):
                actual = ODataUtils.req_svc_products(odata, count=True)
        self.assertEqual((data['value'], data['@odata.count']), actual)

    @patch('drb.drivers.odata.odata_node.OdataNode')
    def test_req_product_by_uuid(self, odata):
        expected = {
            'Id': str(uuid.uuid4()),
            'Name': 'foobar',
        }

        with io.BytesIO(json.dumps(expected).encode()) as resp:
            with patch.object(DrbHttpNode, 'get_impl', return_value=resp):
                actual = ODataUtils.req_product_by_uuid(odata, expected['Id'])

        self.assertEqual(expected, actual)

    @patch('drb.drivers.odata.odata_node.OdataNode')
    def test_req_product_download(self, odata):
        prd_id = str(uuid.uuid4())
        expected = b'downloading_data'

        odata.get_service_url.return_value = 'https://my.domain.com'
        odata.get_auth.return_value = None
        type(odata).type_service = unittest.mock.PropertyMock(
            return_value=ODataServiceType.CSC)

        with io.BytesIO(expected) as resp:
            with patch.object(DrbHttpNode, 'get_impl', return_value=resp):
                actual = ODataUtils.req_product_download(odata, prd_id)
            self.assertEqual(expected, actual.read())


class TestODataQueryPredicate(unittest.TestCase):
    def test_matches(self):
        self.assertFalse(ODataQueryPredicate().matches(DrbLogicalNode('foo')))



import unittest

from drb.exceptions.core import DrbException

from drb.drivers.odata import ODataServiceNodeCSC, ODataProductNode
from tests.utils import start_mock_odata_csc, stop_mock_odata_csc


class TestAttributeNode(unittest.TestCase):
    service_url = 'https://gael-systems.com/csc'
    uuid = '0723d9a4-3bbe-305e-b712-5e820058e065'
    node = None

    @classmethod
    def setUpClass(cls) -> None:
        start_mock_odata_csc(cls.service_url)
        cls.node = ODataServiceNodeCSC(cls.service_url)[0]
        cls.node = cls.node.get_impl(ODataProductNode)

    @classmethod
    def tearDownClass(cls) -> None:
        stop_mock_odata_csc()

    def test_name(self):
        self.assertEqual('lineage', self.node[0][0].name)
        self.assertEqual('instrumentShortName', self.node[0][2].name)

    def test_namespace_uri(self):
        for i in range(49):
            self.assertIsNone(self.node[0][i].namespace_uri)

    def test_value(self):
        self.assertEqual('MSI', self.node[0][2].value)
        self.assertEqual(-13.5554284611837, self.node[0][16].value)
        self.assertEqual(23371, self.node[0][33].value)

    def test_path(self):
        self.assertEqual(f'{self.node.path.name}/Attributes/lineage',
                         self.node[0][0].path.name)
        self.assertEqual(f'{self.node.path.name}'
                         f'/Attributes/instrumentShortName',
                         self.node[0][2].path.name)

    def test_parent(self):
        for i in range(49):
            self.assertEqual(self.node[0], self.node[0][i].parent)

    def test_attributes(self):
        attributes = self.node[0][25].attributes
        self.assertEqual('endingDateTime', attributes['Name', None])
        with self.assertRaises(KeyError):
            attributes['test_name', 'test_ns']

    def test_children(self):
        for i in range(49):
            self.assertEqual([], self.node[0][i].children)

    def test_get_attribute(self):
        self.assertEqual('endingDateTime',
                         self.node[0][25].get_attribute('Name'))
        self.assertEqual('DateTimeOffset',
                         self.node[0][25].get_attribute('ValueType'))
        self.assertEqual('2019-12-13T10:51:40Z',
                         self.node[0][25].get_attribute('Value'))
        with self.assertRaises(DrbException):
            self.node[0].get_attribute('test')

    def test_has_child(self):
        for i in range(49):
            self.assertFalse(self.node[0][i].has_child())

    def test_has_impl(self):
        for i in range(49):
            self.assertFalse(self.node[0][i].has_impl(int))

    def test_get_impl(self):
        for i in range(49):
            with self.assertRaises(DrbException):
                self.node[0][i].get_impl(str)



from __future__ import annotations

import uuid
import datetime
from abc import ABC, abstractmethod

from typing import Any, Dict, Iterable, \
    List, Optional, Tuple, Union

from drb.core import ParsedPath, DrbNode, Predicate
from drb.exceptions.core import DrbException, DrbFactoryException
from drb.topics import resolver
from requests.auth import AuthBase

from .odata_node import OdataNode
from .odata_nodes import ODataOrderNode, ODataProductNode
from .odata_utils import ODataUtils, ODataServiceType, ODataQueryPredicate


class ODataServiceNode(OdataNode, ABC):
    def __init__(self, url: str, auth: AuthBase = None):
        super(ODataServiceNode, self).__init__(url, auth)
        self.__path = ParsedPath(url)
        self._type = None

    @property
    def type_service(self) -> ODataServiceType:
        raise NotImplementedError

    @property
    def orders(self):
        raise NotImplementedError

    @property
    def orders_eta(self):
        raise NotImplementedError

    @property
    def name(self) -> str:
        return self._service_url

    @property
    def value(self) -> Optional[Any]:
        return None

    @property
    def id(self):
        raise NotImplementedError

    @property
    def order(self) -> str:
        raise NotImplementedError

    @property
    def path(self) -> ParsedPath:
        return self.__path

    @property
    def parent(self) -> Optional[DrbNode]:
        return None

    @property
    def attributes(self) -> Dict[Tuple[str, str], Any]:
        return {}

    @property
    def children(self) -> List[DrbNode]:
        filtered_list = ODataFilteredList(self)
        children_list = ODataServiceNodeList(filtered_list)
        return children_list

    def get_attribute(self, name: str, namespace_uri: str = None) -> Any:
        raise DrbException('ODataNode has no attribute')

    def has_child(self, name: str = None, namespace: str = None) -> bool:
        if namespace is None:
            if name is not None:
                try:
                    self._get_named_child(name=name)
                except DrbException:
                    return False
                return True
            return ODataUtils.req_svc_count(self) > 0
        return False

    def close(self) -> None:
        pass

    def has_impl(self, impl: type) -> bool:
        return False

    def get_impl(self, impl: type, **kwargs) -> Any:
        raise DrbException(f"ODataNot doesn't support {impl} implementation")

    @abstractmethod
    def prepare_filter(self, user_filter: str) -> str:
        raise NotImplementedError

    @abstractmethod
    def _get_named_child(self, name: str, namespace_uri: str = None,
                         occurrence: Union[int, slice] = 0) \
            -> Union[DrbNode, List[DrbNode]]:
        raise NotImplementedError

    def __retrieve_child_from_tuple(self, t: tuple) -> \
            Union[List[DrbNode], DrbNode]:
        if len(t) == 2:
            # (name, namespace)
            if t[1] is None or isinstance(t[1], str):
                return self._get_named_child(name=t[0], namespace_uri=t[1])
            # (name, occurrence)
            elif isinstance(t[1], int) or isinstance(t[1], slice):
                return self._get_named_child(t[0], occurrence=t[1])
        # (name, namespace, occurrence)
        elif len(t) == 3:
            return self._get_named_child(*t)
        raise KeyError(f'Invalid key: {t}')

    @abstractmethod
    def retrieve_default_criteria(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def generate_product(self, data: dict) -> ODataProductNode:
        raise NotImplementedError

    def __retrieve_child(self, key: Union[uuid.UUID, tuple, Predicate]) -> \
            Union[List[DrbNode], DrbNode]:
        try:
            if isinstance(key, uuid.UUID):
                prd_uuid = str(key)
                data = ODataUtils.req_product_by_uuid(self,
                                                      self.format_product(
                                                          prd_uuid
                                                      ))
                return ODataProductNode(self, str(prd_uuid), data=data)
            if isinstance(key, tuple):
                return self.__retrieve_child_from_tuple(key)
            if isinstance(key, str):
                return self._get_named_child(key)
            if isinstance(key, ODataQueryPredicate):
                return ODataFilteredList(self, qfilter=key.filter,
                                         qorder=key.order)
        except DrbException as ex:
            raise KeyError from ex
        raise TypeError('Invalid type for a DrbNode bracket: '
                        f'{key.__class__}')

    def __len__(self):
        return ODataUtils.req_svc_count(self)

    def __getitem__(self, item):
        if isinstance(item, int):
            children_list = self.children
            resolved_node = children_list[item]
            return resolved_node

        elif isinstance(item, uuid.UUID):
            node = self.__retrieve_child(item)
            try:
                if node.is_online:
                    _, resolved_node = resolver.resolve(
                        node
                    )
                    return resolved_node
                else:
                    return node
            except DrbFactoryException:
                return node

        elif isinstance(item, ODataQueryPredicate):
            filtered_list = self.__retrieve_child(item)
            children_list = ODataServiceNodeList(filtered_list)
            return children_list

        elif isinstance(item, str):
            node = self.__retrieve_child(item)
            try:
                _, resolved_node = resolver.resolve(node)
                return resolved_node
            except DrbFactoryException:
                return node

        elif isinstance(item, tuple):
            node = self.__retrieve_child(item)
            try:
                _, resolved_node = resolver.resolve(node)
                return resolved_node
            except DrbFactoryException:
                return node

        elif isinstance(item, slice):
            children_list = self.children
            sliced_list = children_list[item]
            return sliced_list

        else:
            raise TypeError

    def format_product(self, uuid: str):
        raise NotImplementedError

    @property
    def format_attributes(self):
        raise NotImplementedError

    def format_order(self, uuid: str):
        raise NotImplementedError

    @property
    def filter_keyword(self) -> str:
        raise NotImplementedError

    @property
    def is_count_accepted_in_request(self) -> bool:
        raise NotImplementedError


class ODataServiceNodeCSC(ODataServiceNode):

    @property
    def id(self):
        return 'Id'

    @property
    def order(self) -> str:
        return 'OData.CSC.Order'

    @property
    def orders_eta(self) -> str:
        return 'EstimatedDate'

    @property
    def filter_keyword(self) -> str:
        return '$filter'

    @property
    def is_count_accepted_in_request(self) -> bool:
        return True

    def format_order(self, uuid: str):
        return "Orders({0})".format(uuid)

    def format_product(self, uuid: str):
        return "Products({0})".format(uuid)

    def format_attributes(self):
        return "Attributes"

    @property
    def type_service(self) -> ODataServiceType:
        self._type = ODataServiceType.CSC
        return self._type

    def retrieve_default_criteria(self) -> str:
        return 'PublicationDate'

    def prepare_filter(self, user_filter: str) -> str:
        date = datetime.datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%SZ')
        if user_filter is None:
            return f'{self.retrieve_default_criteria()} lt {date}'
        if self.retrieve_default_criteria() in user_filter:
            return user_filter
        return f"{user_filter} and" \
               f" {self.retrieve_default_criteria()} lt {date}"

    @property
    def namespace_uri(self) -> Optional[str]:
        return 'OData.CSC'

    def _get_named_child(self, name: str,
                         namespace_uri: str = None,
                         occurrence: Union[int, slice] = 0
                         ) -> Union[DrbNode, List[DrbNode]]:
        if namespace_uri is None:
            products = ODataUtils.req_svc_products(
                self, filter=f"Name eq '{name}'")
            if len(products) <= 0:
                raise DrbException(f'No child ({name}, {namespace_uri},'
                                   f' {occurrence}) found')
            p = products[occurrence]
            if isinstance(p, list):
                return [ODataProductNode(self,
                                         x['Id'], data=x) for x in p]
            return ODataProductNode(self, p['Id'], data=p)
        raise DrbException(f'No child ({name}, {namespace_uri}, {occurrence}) '
                           'found')

    @property
    def orders(self):
        return OdataOrderList(self)

    def generate_product(self, data: dict) -> ODataProductNode:
        return ODataProductNode(self, data['Id'], data=data)


class ODataServiceNodeDhus(ODataServiceNode):

    @property
    def orders(self):
        raise NotImplementedError

    @property
    def id(self):
        return 'Id'

    @property
    def order(self) -> str:
        raise NotImplementedError

    @property
    def orders_eta(self) -> str:
        raise NotImplementedError

    @property
    def filter_keyword(self) -> str:
        return '$filter'

    @property
    def is_count_accepted_in_request(self) -> bool:
        return True

    def format_order(self, uuid: str):
        return "Orders('{0}')".format(uuid)

    def format_product(self, uuid: str):
        return "Products('{0}')".format(uuid)

    def format_attributes(self):
        return "Attributes"

    @property
    def type_service(self) -> ODataServiceType:
        self._type = ODataServiceType.DHUS
        return self._type

    def retrieve_default_criteria(self) -> str:
        return 'CreationDate'

    def prepare_filter(self, user_filter: str) -> str:
        date = datetime.datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%S.0Z')
        if user_filter is None:
            return f'{self.retrieve_default_criteria()} lt {date}'
        if self.retrieve_default_criteria() in user_filter:
            return user_filter
        return f"{user_filter} and " \
               f"{self.retrieve_default_criteria()} lt {date}"

    @property
    def children(self) -> List[DrbNode]:
        return OdataFilteredListDhus(self, page_size=10)

    @property
    def namespace_uri(self) -> Optional[str]:
        return 'OData.DHuS'

    def _get_named_child(self, name: str,
                         namespace_uri: str = None,
                         occurrence: Union[int, slice] = 0
                         ) -> Union[DrbNode, List[DrbNode]]:
        if namespace_uri is None:
            products = ODataUtils.req_svc_products(
                self, filter=f"Name eq '{name}'")
            if len(products) <= 0:
                raise DrbException(f'No child ({name}, {namespace_uri},'
                                   f' {occurrence}) found')
            p = products[occurrence]
            if isinstance(p, list):
                return [ODataProductNode(self,
                                         x['Id'], data=x) for x in p]
            return ODataProductNode(self, p['Id'], data=p)
        raise DrbException(f'No child ({name}, {namespace_uri}, {occurrence}) '
                           'found')

    def __retrieve_child(self, key: Union[uuid.UUID, tuple, Predicate]) -> \
            Union[List[DrbNode], DrbNode]:
        try:
            if isinstance(key, uuid.UUID):
                prd_uuid = str(key)
                data = ODataUtils.req_product_by_uuid(self,
                                                      self.format_product(
                                                          prd_uuid
                                                      ))
                return ODataProductNode(self, str(prd_uuid), data=data)
            if isinstance(key, tuple):
                return self.__retrieve_child_from_tuple(key)
            if isinstance(key, str):
                return self._get_named_child(key)
            if isinstance(key, ODataQueryPredicate):
                return OdataFilteredListDhus(
                    self,
                    qfilter=key.filter,
                    qorder=key.order)
        except DrbException as ex:
            raise KeyError from ex
        raise TypeError('Invalid type for a DrbNode bracket: '
                        f'{key.__class__}')

    def __getitem__(self, item):
        if isinstance(item, int):
            children_list = self.children
            resolved_node = children_list[item]
            return resolved_node

        elif isinstance(item, uuid.UUID):
            node = self.__retrieve_child(item)
            try:
                if node.is_online:
                    _, resolved_node = resolver.resolve(
                        node
                    )
                    return resolved_node
                else:
                    return node
            except DrbFactoryException:
                return node

        elif isinstance(item, ODataQueryPredicate):
            filtered_list = self.__retrieve_child(item)
            children_list = ODataServiceNodeList(filtered_list)
            return children_list

        elif isinstance(item, str):
            node = self.__retrieve_child(item)
            try:
                _, resolved_node = resolver.resolve(node)
                return resolved_node
            except DrbFactoryException:
                return node

        elif isinstance(item, tuple):
            node = self.__retrieve_child(item)
            try:
                _, resolved_node = resolver.resolve(node)
                return resolved_node
            except DrbFactoryException:
                return node

        elif isinstance(item, slice):
            children_list = self.children
            sliced_list = children_list[item]
            return sliced_list

        else:
            raise TypeError

    def generate_product(self, data: dict) -> ODataProductNode:
        return ODataProductNode(self, data['Id'], data=data)


class ODataServiceNodeDias(ODataServiceNode):

    @property
    def orders(self):
        return OdataOrderList(self)

    @property
    def id(self):
        return 'id'

    @property
    def order(self) -> str:
        return 'Ens.Order'

    @property
    def orders_eta(self) -> str:
        return 'EstimatedTime'

    @property
    def filter_keyword(self) -> str:
        return '$search'

    @property
    def is_count_accepted_in_request(self) -> bool:
        return False

    def format_order(self, uuid: str):
        return "Orders('{0}')".format(uuid)

    def format_product(self, uuid: str):
        return "Products({0})".format(uuid)

    def format_attributes(self):
        return "Metadata"

    @property
    def type_service(self) -> ODataServiceType:
        self._type = ODataServiceType.ONDA_DIAS
        return self._type

    def retrieve_default_criteria(self) -> str:
        return 'creationDate'

    def prepare_filter(self, user_filter: str) -> str:
        if user_filter is None:
            return f"%22creationDate:[1900-05-10T00:00:00.000Z TO NOW]%22"
        return f"%22{user_filter}%22"

    @property
    def namespace_uri(self) -> Optional[str]:
        return 'Ens'

    def _get_named_child(self, name: str, namespace_uri: str = None,
                         occurrence: Union[int, slice] = 0
                         ) -> Union[DrbNode, List[DrbNode]]:
        if namespace_uri is None:
            products = ODataUtils.req_svc_products(
                self, search=f'"name:{name}"')
            if len(products) <= 0:
                raise DrbException(f'No child ({name}, {namespace_uri},'
                                   f' {occurrence}) found')
            p = products[occurrence]
            if isinstance(p, list):
                return [ODataProductNode(self,
                                         x['id'], data=x) for x in p]
            return ODataProductNode(self, p['id'], data=p)
        raise DrbException(f'No child ({name}, {namespace_uri}, {occurrence}) '
                           'found')

    def generate_product(self, data: dict) -> ODataProductNode:
        return ODataProductNode(self, data['id'], data=data)


class ODataFilteredList(list):

    def __init__(self, odata: OdataNode, qfilter: str = None,
                 qorder: str = None, page_size=100):
        super().__init__()
        self._count = -1
        self._page = None
        self._skip = 0
        self._page_size = page_size
        self._odata = odata
        default = odata.retrieve_default_criteria()
        self._filter = odata.prepare_filter(qfilter)
        self._order = self.__prepare_order(qorder, default)

    @classmethod
    def __prepare_order(cls, user_order: str, default: str) -> str:
        if user_order is None:
            return f'{default} desc'
        return user_order

    def __perform_query(self):
        buffer, self._count = ODataUtils.req_svc_products(
            self._odata, count=self._count, filter=self._filter,
            order=self._order, skip=self._skip, top=self._page_size)

        if self._count == -1 and not self._odata.is_count_accepted_in_request:
            self._count = ODataUtils.req_svc_count_search(self._odata,
                                                          self._filter)

        self._page = [self._odata.generate_product(e) for e in buffer]

    def __compute_index(self, item: Union[int, slice]) -> Tuple[int, int]:
        if isinstance(item, int):
            if item < 0:
                return item + len(self), -1
            return item, -1
        # item is a slice
        start = item.start if item.start is not None else 0
        if start < 0:
            start = start + len(self)
        stop = item.stop if item.stop is not None else len(self)
        if stop < 0:
            stop = stop + len(self)
        return start, stop

    def append(self, obj: Any) -> None:
        raise NotImplementedError

    def clear(self) -> None:
        raise NotImplementedError

    def copy(self) -> List[ODataProductNode]:
        raise NotImplementedError

    def count(self, value: Any) -> int:
        raise NotImplementedError

    def extend(self, iterable: Iterable[ODataProductNode]) -> None:
        raise NotImplementedError

    def index(self, value: Any, start: int = ..., stop: int = ...) -> int:
        raise NotImplementedError

    def insert(self, index: int, obj: ODataProductNode) -> None:
        raise NotImplementedError

    def pop(self, index: int = ...) -> ODataProductNode:
        raise NotImplementedError

    def remove(self, value: Any) -> None:
        raise NotImplementedError

    def reverse(self) -> None:
        raise NotImplementedError

    def sort(self: List, *, key: None = ..., reverse: bool = ...) -> None:
        raise NotImplementedError

    def __getitem_index(self, index):
        if index not in range(self._skip, self._skip + self._page_size):
            page_number = index // self._page_size
            self._skip = page_number * self._page_size
            self.__perform_query()
        return self._page[index % self._page_size]

    def __getitem__(self, item):
        if self._count == -1:
            self.__perform_query()

        if isinstance(item, int):
            index, _ = self.__compute_index(item)
            return self.__getitem_index(index)
        elif isinstance(item, slice):
            start, stop = self.__compute_index(item)
            result = []
            for index in range(start, stop):
                node = self.__getitem_index(index)
                result.append(node)
            return result
        else:
            raise KeyError(f'Invalid key: {type(item)}')

    def __iter__(self):
        def generator(qfilter, qorder, count, page_size):
            page_num = 0
            buffer = None
            for i in range(count):
                if i % page_size == 0:
                    skip = page_size * page_num
                    buffer = ODataUtils.req_svc_products(
                        self._odata, filter=qfilter, order=qorder, skip=skip,
                        top=page_size)
                    page_num += 1
                idx = i % page_size
                yield self._odata.generate_product(buffer[idx])

        return generator(self._filter, self._order, len(self), self._page_size)

    def __len__(self):
        if self._count == -1:
            self.__perform_query()
        return self._count

    def __contains__(self, item):
        raise NotImplementedError

    def __iadd__(self, other):
        raise NotImplementedError

    def __imul__(self, other):
        raise NotImplementedError


class OdataFilteredListDhus(ODataFilteredList):

    def __init__(self, odata: OdataNode, qfilter: str = None,
                 qorder: str = None, page_size=100):
        self._is_filtered = qfilter
        super().__init__(
            odata,
            qfilter,
            qorder,
            page_size)

    def __len__(self):
        if self._is_filtered is not None:
            raise IndexError('Count is not supported on this '
                             'service with filtered request.')
        if self._count == -1:
            self.__perform_query()
        return self._count

    def __perform_query(self):
        buffer = ODataUtils.req_svc_products(
            self._odata, filter=self._filter,
            order=self._order, skip=self._skip, top=self._page_size)

        if self._count == -1 and not self._odata.is_count_accepted_in_request \
                and self._is_filtered is None:
            self._count = ODataUtils.req_svc_count_search(self._odata,
                                                          self._filter)
        elif self._count == -1:
            self._count = 0

        self._count += self._page_size

        self._page = [self._odata.generate_product(e) for e in buffer]

    def __compute_index(self, item: Union[int, slice]) -> Tuple[int, int]:
        if isinstance(item, int):
            if item < 0:
                return item + self._count, -1
            return item, -1
        # item is a slice
        start = item.start if item.start is not None else 0
        if start < 0:
            start = start + self._count
        stop = item.stop if item.stop is not None else len(self)
        if stop < 0:
            stop = stop + self._count
        return start, stop

    def __iter__(self):
        def generator(qfilter, qorder, count, page_size):
            page_num = 0
            buffer = None
            for i in range(count):
                if i % page_size == 0:
                    skip = page_size * page_num
                    buffer = ODataUtils.req_svc_products(
                        self._odata, filter=qfilter, order=qorder, skip=skip,
                        top=page_size)
                    page_num += 1
                idx = i % page_size
                yield self._odata.generate_product(buffer[idx])

        return generator(self._filter, self._order,
                         self._count, self._page_size)


class OdataOrderList(list):
    def __init__(self, odata: ODataServiceNode, qfilter: str = None,
                 page_size=100):
        super().__init__()
        self._count = -1
        self._page = None
        self._skip = 0
        self._page_size = page_size
        self._filter = qfilter
        self._odata = odata

    def __perform_query(self):
        buffer, self._count = ODataUtils.req_svc_orders(
            self._odata, count=self._count,
            filter=self._filter,
            skip=self._skip,
            top=self._page_size)

        self._page = [ODataOrderNode(self._odata, e) for e in buffer]

    def __compute_index(self, item: Union[int, slice]) -> Tuple[int, int]:
        if isinstance(item, int):
            if item < 0:
                return item + len(self), -1
            return item, -1
        # item is a slice
        start = item.start if item.start is not None else 0
        if start < 0:
            start = start + len(self)
        stop = item.stop if item.stop is not None else len(self)
        if stop < 0:
            stop = stop + len(self)
        return start, stop

    def append(self, obj: Any) -> None:
        raise NotImplementedError

    def clear(self) -> None:
        raise NotImplementedError

    def copy(self) -> List[ODataProductNode]:
        raise NotImplementedError

    def count(self, value: Any) -> int:
        raise NotImplementedError

    def extend(self, iterable: Iterable[ODataProductNode]) -> None:
        raise NotImplementedError

    def index(self, value: Any, start: int = ..., stop: int = ...) -> int:
        raise NotImplementedError

    def insert(self, index: int, obj: ODataProductNode) -> None:
        raise NotImplementedError

    def pop(self, index: int = ...) -> ODataProductNode:
        raise NotImplementedError

    def remove(self, value: Any) -> None:
        raise NotImplementedError

    def reverse(self) -> None:
        raise NotImplementedError

    def sort(self: List, *, key: None = ..., reverse: bool = ...) -> None:
        raise NotImplementedError

    def __getitem_index(self, index):
        if index not in range(self._skip, self._skip + self._page_size):
            page_number = index // self._page_size
            self._skip = page_number * self._page_size
            self.__perform_query()
        return self._page[index % self._page_size]

    def __getitem__(self, item):
        if isinstance(item, uuid.UUID):
            order_uuid = str(item)
            data = ODataUtils.get_order_by_uuid(self._odata,
                                                self._odata.format_order(
                                                    order_uuid
                                                ))
            return ODataOrderNode(self._odata, source=data)

        if self._count == -1:
            self.__perform_query()

        if isinstance(item, int):
            index, _ = self.__compute_index(item)
            return self.__getitem_index(index)
        elif isinstance(item, slice):
            start, stop = self.__compute_index(item)
            result = []
            for index in range(start, stop):
                node = self.__getitem_index(index)
                result.append(node)
            return result
        elif isinstance(item, ODataQueryPredicate):
            return OdataOrderList(self._odata, qfilter=item.filter)
        else:
            raise KeyError(f'Invalid key: {type(item)}')

    def __iter__(self):
        def generator(count, page_size):
            page_num = 0
            buffer = None
            for i in range(count):
                if i % page_size == 0:
                    skip = page_size * page_num
                    buffer = ODataUtils.req_svc_orders(
                        self._odata, skip=skip,
                        top=page_size)
                    page_num += 1
                idx = i % page_size
                yield ODataOrderNode(self._odata, buffer[idx])

        return generator(len(self), self._page_size)

    def __len__(self):
        if self._count == -1:
            self.__perform_query()
        return self._count

    def __contains__(self, item):
        raise NotImplementedError

    def __iadd__(self, other):
        raise NotImplementedError

    def __imul__(self, other):
        raise NotImplementedError


class ODataServiceNodeList(list):

    def __init__(self, odata: Union[List[DrbNode], ODataFilteredList]):
        super().__init__()
        self._odata = odata

    def __getitem__(self, item):

        if isinstance(item, slice):
            sliced_list = self._odata[item]
            result = ODataServiceNodeList(sliced_list)
            return result

        elif isinstance(item, int):
            node = self._odata[item]
            try:
                if node.is_online:
                    _, resolved_node = resolver.resolve(
                        node
                    )
                    return resolved_node
                else:
                    return node
            except DrbFactoryException:
                return node
        else:
            raise KeyError(f'Invalid key: {type(item)}')

    def __iter__(self):
        def gen():
            i = 0
            while 1:
                try:
                    node = self._odata[i]
                    try:
                        _, resolved_node = resolver.resolve(node)
                    except DrbFactoryException:
                        return node
                    yield resolved_node
                    i += 1
                except IndexError:
                    break

        return gen()

    def __len__(self):
        count = len(self._odata)
        return count

    def append(self, obj: Any) -> None:
        raise NotImplementedError

    def clear(self) -> None:
        raise NotImplementedError

    def copy(self) -> List[ODataProductNode]:
        raise NotImplementedError

    def count(self, value: Any) -> int:
        raise NotImplementedError

    def extend(self, iterable: Iterable[ODataProductNode]) -> None:
        raise NotImplementedError

    def index(self, value: Any, start: int = ..., stop: int = ...) -> int:
        raise NotImplementedError

    def insert(self, index: int, obj: ODataProductNode) -> None:
        raise NotImplementedError

    def pop(self, index: int = ...) -> ODataProductNode:
        raise NotImplementedError

    def remove(self, value: Any) -> None:
        raise NotImplementedError

    def reverse(self) -> None:
        raise NotImplementedError

    def sort(self: List, *, key: None = ..., reverse: bool = ...) -> None:
        raise NotImplementedError



from __future__ import annotations

import io
import json
import time
from multiprocessing import Process
from deprecated import deprecated

from typing import Any, Dict, List, Optional, Tuple, Union

from drb.core import ParsedPath, DrbNode
from drb.exceptions.core import DrbException, DrbNotImplementationException
from drb.nodes.abstract_node import AbstractNode

from .odata_node import OdataNode
from .odata_utils import ODataUtils, ODataServiceType
from ...exceptions.odata import OdataRequestException

process = None


class ODataProductNode(OdataNode):
    def __init__(self, source: Union[OdataNode, str], product_uuid: str = None,
                 **kwargs):
        self.__initialized = False
        self.__name = None
        if isinstance(source, OdataNode):
            super().__init__(source.get_service_url(), source.get_auth())
            self.parent = source
            self.__uuid = product_uuid
        elif isinstance(source, str):
            svc_prd_id = source.split('/Products')
            auth = kwargs.get('auth', None)
            super().__init__(svc_prd_id[0], auth)
            self.__type = ODataUtils.get_type_odata_svc(svc_prd_id[0], auth)
            self.__uuid = product_uuid
        if 'data' in kwargs:
            data = kwargs['data']
            self.__uuid = data.get('Id', data.get('id'))
            self.__product = data
            self.name = data.get('Name', data.get('name'))
        self.__path = ParsedPath(
            f'{self.get_service_url()}/Products({self.__uuid})')
        self._available_impl.append(ODataProductNode)

    def format_product(self, uuid: str):
        if self.parent is not None:
            return self.parent.format_product(self.__uuid)
        if self.type_service == ODataServiceType.DHUS:
            return "Products('{0}')".format(uuid)
        return "Products({0})".format(uuid)

    @property
    def type_service(self) -> ODataServiceType:
        if self.parent is not None:
            return self.parent.type_service
        elif not hasattr(self, '__type'):
            self.__type = ODataUtils.get_type_odata_svc(self.get_service_url())
        return self.__type

    def __load_product(self):
        if not hasattr(self, ' __product'):
            self.__product = ODataUtils.req_product_by_uuid(
                self,
                self.format_product(self.__uuid))

    @property
    def name(self) -> str:
        if self.__name is None:
            self.__load_product()
            if self.type_service == ODataServiceType.ONDA_DIAS:
                self.__name = self.__product['name']
            else:
                self.__name = self.__product['Name']
        return self.__name

    @name.setter
    def name(self, value: str):
        self.__name = value

    @property
    def path(self) -> ParsedPath:
        return self.__path

    @property
    @deprecated(version='2.1.0')
    def children(self) -> List[DrbNode]:
        return [ODataProductAttributeNode(self, self.__uuid)]

    # FIXME: implemented here to ensure a complete initialization of the node
    @deprecated(version='2.1.0')
    def get_attribute(self, name: str, namespace_uri: str = None) -> Any:
        return self @ (name, namespace_uri)

    def order(self):
        if self.type_service in [
            ODataServiceType.CSC, ODataServiceType.ONDA_DIAS] and \
                self.is_online is False:
            resp = ODataUtils.launch_order(self.parent,
                                           self.get_attribute(self.parent.id),
                                           self.parent.order)
            order = ODataOrderNode(self.parent, json.loads(
                resp.value.decode()))

            return order
        else:
            raise NotImplemented()

    def has_impl(self, impl: type) -> bool:
        if impl == ODataProductNode:
            return True
        if self.type_service == ODataServiceType.ONDA_DIAS:
            online = not self.get_attribute('offline')
        else:
            online = self.get_attribute('Online')
        return online and (impl == io.BufferedIOBase or impl == io.BytesIO)

    def get_impl(self, impl: type, **kwargs) -> Any:
        if self.has_impl(impl):
            if impl == ODataProductNode:
                return self
            return ODataUtils.req_product_download(self, self.format_product(
                self.__uuid),
                                                   kwargs.get('start', None),
                                                   kwargs.get('end', None))
        raise DrbException(f'Not supported implementation: {impl}')

    def __eq__(self, other):
        if isinstance(other, ODataProductNode):
            return OdataNode.__eq__(self, other) and \
                self.__uuid == other.__uuid
        return False

    def __hash__(self):
        return hash(self._service_url)

    def __matmul__(self, other):
        self.__load_product()
        if len(self.attribute_names()) == 0:
            for k, v in self.__product.items():
                if v is not None:
                    self @= (k, None, v)
        return super().__matmul__(other)

    @property
    def is_online(self):
        if not self.attribute_names():
            self.__load_product()
        if self.type_service in [ODataServiceType.DHUS, ODataServiceType.CSC]:
            return self @ 'Online'
        elif self.type_service == ODataServiceType.ONDA_DIAS:
            return not self @ 'offline'
        return None


class ODataProductAttributeNode(OdataNode):
    __name = 'Attributes'

    def __init__(self, source: ODataProductNode, prd_uuid: str):
        super().__init__(source.get_service_url(), source.get_auth())
        self.name = self.__name
        self.parent = source
        self.__uuid = prd_uuid
        self.__attr = None
        self._available_impl.clear()

    @property
    def type_service(self) -> ODataServiceType:
        return self.parent.type_service

    def __load_attributes(self) -> None:
        if self.__attr is None:
            self.__attr = ODataUtils.req_product_attributes(
                self,
                self.parent.parent.format_product(self.__uuid),
                self.parent.parent.format_attributes()
            )

    @property
    def path(self) -> ParsedPath:
        return self.parent.path / self.__name

    @property
    def children(self) -> List[DrbNode]:
        self.__load_attributes()
        return [ODataAttributeNode(self, data=x) for x in self.__attr]

    def has_child(self, name: str = None, namespace: str = None) -> bool:
        if namespace is None:
            if name is not None:
                for x in self.__attr:
                    if name in x.keys():
                        return True
                return False
            return len(self.__attr) > 0
        return False

    def get_impl(self, impl: type, **kwargs) -> Any:
        raise DrbException(f'Do not support implementation: {impl}')


class ODataAttributeNode(OdataNode):
    def __init__(self, source: Union[str, ODataProductAttributeNode],
                 **kwargs):
        if isinstance(source, ODataProductAttributeNode):
            super().__init__(source.get_service_url(), source.get_auth())
            self.parent = source
            self._type = source.type_service
        elif isinstance(source, str):
            auth = kwargs.get('auth')
            super().__init__(source, auth)
            self._type = ODataServiceType.UNKNOWN
        else:
            raise OdataRequestException(f'Unsupported source: {type(source)}')
        self.__path = None
        self.__data = kwargs.get('data')
        for k, v in self.__data.items():
            self @= (k, None, v)
        self.name = self.__data.get('Name', self.__data.get('name'))
        self.value = self.__data.get('Value', self.__data.get('value'))
        self._available_impl.clear()

    @property
    def type_service(self) -> ODataServiceType:
        if self.parent is None:
            return ODataServiceType.UNKNOWN
        return self.parent.type_service

    @property
    def path(self) -> ParsedPath:
        if self.__path is None:
            if self.parent is None:
                self.__path = ParsedPath(self.name)
            else:
                self.__path = self.parent.path / self.name
        return self.__path

    @property
    def children(self) -> List[DrbNode]:
        return []

    def get_impl(self, impl: type, **kwargs) -> Any:
        raise DrbException(f'Do not support implementation: {impl}')


class ODataOrderNode(AbstractNode):

    def __init__(self, node: OdataNode, source: dict):
        super().__init__()
        self.name = source['Id']
        self.parent = node
        self.__path = ParsedPath(
            f'{node.get_service_url()}/{node.format_order(self.name)}')
        self._init_attributes(source)

    def __setitem__(self, key, value):
        raise NotImplementedError

    def __delitem__(self, key):
        raise NotImplementedError

    def _init_attributes(self, data):
        for k, v in data.items():
            if v is not None:
                self @= (k, None, v)

    @property
    def product(self):
        data = ODataUtils.req_product_order_by_uuid(self)
        return ODataProductNode(self.parent, data=data)

    @property
    def order_eta(self):
        return self.get_attribute(self.parent.orders_eta)

    @property
    def path(self) -> ParsedPath:
        return self.__path

    @property
    def children(self) -> List[DrbNode]:
        return []

    def get_impl(self, impl: type, **kwargs) -> Any:
        raise DrbNotImplementationException

    @property
    def status(self):
        data = ODataUtils.get_order_by_uuid(
            self.parent,
            self.parent.format_order(self.name)
        )
        self._init_attributes(data)
        return self @ 'Status'

    def cancel(self):
        if self.parent.type_service == ODataServiceType.CSC:
            return ODataUtils.cancel_order(
                self.parent, self.parent.format_order(self.name)
            )
        else:
            raise NotImplemented

    def _wait_and_check(self, order: ODataOrderNode, wait: int = 60):
        while 1:
            if not order.status != 'in_progress':
                break
            res = ODataUtils.get_order_by_uuid(
                self.parent,
                self.parent.format_order(self.name)
            )
            if res['Status'] == 'completed':
                self.product.attributes[('Online', None)] = 'True'
                break
            if res['Status'] in ['failed', 'cancelled']:
                break
            time.sleep(wait)

    def wait(self, step: int = 60):
        global process
        process = Process(target=self._wait_and_check, args=(self, step))
        process.start()
        return True

    def stop(self):
        global process
        process.terminate()



import unittest
import uuid

from drb.core.factory import FactoryLoader
from drb.topics.topic import TopicCategory
from drb.topics.dao import ManagerDao
from drb.nodes.logical_node import DrbLogicalNode
from drb.drivers.odata import OdataFactory
from tests.utils import start_mock_odata_csc, stop_mock_odata_csc


class TestOdataSignature(unittest.TestCase):
    svc_url = 'http+odata://my.domain.com/csc'
    svc_url_false = 'https://my.domain.com/csc'
    fc_loader = None
    ic_loader = None
    odata_id = uuid.UUID('a32c5d56-409e-11ec-973a-0242ac130003')

    @classmethod
    def setUpClass(cls) -> None:
        start_mock_odata_csc(cls.svc_url)
        cls.fc_loader = FactoryLoader()
        cls.ic_loader = ManagerDao()

    @classmethod
    def tearDownClass(cls) -> None:
        stop_mock_odata_csc()

    def test_impl_loading(self):
        factory_name = 'odata'

        factory = self.fc_loader.get_factory(factory_name)
        self.assertIsNotNone(factory)
        self.assertIsInstance(factory, OdataFactory)

        topic = self.ic_loader.get_drb_topic(self.odata_id)
        self.assertIsNotNone(factory)
        self.assertEqual(self.odata_id, topic.id)
        self.assertEqual('OData Copernicus Space Component', topic.label)
        self.assertIsNone(topic.description)
        self.assertEqual(TopicCategory.PROTOCOL, topic.category)
        self.assertEqual('odata', topic.factory)

    def test_impl_signatures(self):
        topic = self.ic_loader.get_drb_topic(self.odata_id)
        node = DrbLogicalNode(self.svc_url)
        self.assertTrue(topic.matches(node))

        node = DrbLogicalNode(f'{self.svc_url_false}')
        self.assertFalse(topic.matches(node))

        node = DrbLogicalNode(f'http://not.odata.svc')
        self.assertFalse(topic.matches(node))



import unittest
import keyring
from keyring.credentials import SimpleCredential
from unittest.mock import patch
from drb.drivers.http import DrbHttpNode
from drb.nodes.logical_node import DrbLogicalNode
from drb.drivers.odata import OdataFactory, ODataServiceNodeCSC
from tests.utils import start_mock_odata_csc, stop_mock_odata_csc


class TestOdataFactory(unittest.TestCase):
    svc_url = 'https://Odata_test.com'
    filter = '/Products?filter=Online eq false'

    @classmethod
    def setUpClass(cls) -> None:
        start_mock_odata_csc(cls.svc_url)

    @classmethod
    def tearDownClass(cls) -> None:
        stop_mock_odata_csc()

    def test_create(self):
        factory = OdataFactory()
        node = factory.create(self.svc_url)
        self.assertIsNotNone(node)
        self.assertIsInstance(node, ODataServiceNodeCSC)
        node = factory.create(DrbLogicalNode(self.svc_url))
        self.assertIsNotNone(node)
        self.assertIsInstance(node, ODataServiceNodeCSC)

    def test_create_from_http(self):
        factory = OdataFactory()
        node = factory.create(DrbHttpNode(self.svc_url+self.filter))
        self.assertIsNotNone(node)
        self.assertIsInstance(node, list)
        node = factory.create(DrbHttpNode(self.svc_url+'/Products'))
        self.assertIsNotNone(node)
        self.assertIsInstance(node, list)

    def test_create_with_keyring(self):
        auth = SimpleCredential('foo', 'bar')
        with patch.object(keyring, 'get_credential', return_value=auth):
            factory = OdataFactory()
            node = factory.create(self.svc_url)
            self.assertIsNotNone(node)
            self.assertIsInstance(node, ODataServiceNodeCSC)



import abc
import keyring
from drb.nodes.abstract_node import AbstractNode
from requests.auth import AuthBase, HTTPBasicAuth
from typing import Optional


class OdataNode(AbstractNode, abc.ABC):
    """
    Common ODataNode interface
    """

    def __init__(self, service_url, auth: AuthBase = None):
        super(OdataNode, self).__init__()
        self._original_service_url = service_url
        self._service_url = service_url.replace('+odata', '') \
            if '+odata' in service_url else service_url
        self.__auth = auth

    def get_service_url(self) -> str:
        """
        Returns URL of the OData service.
        :returns: string URL representation the OData service
        :rtype: str
        """
        return self._service_url

    def get_auth(self) -> Optional[AuthBase]:
        """
        Returns the associated authentication required to access to the OData
        service.
        :returns: an authentication compatible with requests library.
        :rtype: AuthBase
        """
        if self.__auth is None:
            auth = keyring.get_keyring().get_credential(
                self._original_service_url, None)
            if auth is not None:
                self.__auth = HTTPBasicAuth(auth.username, auth.password)
        return self.__auth

    @property
    @abc.abstractmethod
    def type_service(self):
        raise NotImplementedError

    def __eq__(self, other):
        return isinstance(other, OdataNode) and \
            self._service_url == other._service_url

    def __hash__(self):
        return hash(self._service_url)

    def __setitem__(self, key, value):
        raise NotImplementedError

    def __delitem__(self, key):
        raise NotImplementedError



import io
import json
import keyring

from requests.auth import HTTPBasicAuth
from drb.core import DrbFactory, DrbNode
from drb.drivers.http import DrbHttpNode
from drb.drivers.json import JsonNode
from drb.exceptions.core import DrbFactoryException
from .odata_node import OdataNode
from .odata_nodes import ODataProductNode
from .odata_services_nodes import ODataServiceNodeDhus, \
    ODataServiceNodeDias, ODataServiceNodeCSC
from .odata_utils import ODataServiceType, ODataUtils


class OdataFactory(DrbFactory):
    def _create(self, node: DrbNode) -> DrbNode:
        if isinstance(node, OdataNode):
            return node
        if isinstance(node, DrbHttpNode):
            if '$format=json' not in node.path.original_path:
                if '?' not in node.path.original_path:
                    node._path = node.path.original_path + '?&$format=json'
                else:
                    node._path = node.path.original_path + '&$format=json'
            req = node.get_impl(io.BytesIO).read().decode()
            json_node = JsonNode(json.loads(req))
            return [ODataProductNode(source=node.path.original_path,
                                     auth=node.auth,
                                     data=e.value
                                     ) for e in json_node['value', :]]
        final_url = node.path.name.replace('+odata', '')

        auth = keyring.get_keyring().get_credential(final_url, None)
        if auth is not None:
            auth = HTTPBasicAuth(auth.username, auth.password)

        service_type = ODataUtils.get_type_odata_svc(final_url, auth)
        if service_type == ODataServiceType.CSC:
            return ODataServiceNodeCSC(final_url, auth)
        if service_type == ODataServiceType.DHUS:
            return ODataServiceNodeDhus(final_url, auth)
        if service_type == ODataServiceType.ONDA_DIAS:
            return ODataServiceNodeDias(final_url, auth)

        raise DrbFactoryException(f'Unsupported Odata service: {final_url}')



import unittest

from drb.exceptions.core import DrbException

from drb.drivers.odata import ODataServiceNodeCSC, ODataProductNode
from tests.utils import start_mock_odata_csc, stop_mock_odata_csc


class TestProductAttributeNode(unittest.TestCase):
    service_url = 'https://gael-systems.com/csc'
    uuid = '0723d9a4-3bbe-305e-b712-5e820058e065'
    node = None

    @classmethod
    def setUpClass(cls) -> None:
        start_mock_odata_csc(cls.service_url)
        cls.node = ODataServiceNodeCSC(cls.service_url)[0]
        cls.node = cls.node.get_impl(ODataProductNode)

    @classmethod
    def tearDownClass(cls) -> None:
        stop_mock_odata_csc()

    def test_name(self):
        self.assertEqual('Attributes', self.node[0].name)

    def test_namespace_uri(self):
        self.assertIsNone(self.node[0].namespace_uri)

    def test_value(self):
        self.assertIsNone(self.node[0].value)

    def test_path(self):
        self.assertEqual(
            f'{self.service_url}/Products({self.uuid})/Attributes',
            self.node[0].path.name)

    def test_parent(self):
        self.assertEqual(self.uuid, self.node[0].parent.get_attribute('Id'))

    def test_attribute(self):
        self.assertEqual({}, self.node[0].attributes)

    def test_children(self):
        self.assertIsNotNone(self.node[0].children)
        self.assertIsInstance(self.node[0].children, list)
        self.assertEqual(49, len(self.node[0].children))

    def test_get_attribute(self):
        with self.assertRaises(DrbException):
            self.node[0].get_attribute('fake')

    def test_has_child(self):
        self.assertTrue(self.node.has_child())

    def test_has_impl(self):
        self.assertFalse(self.node[0].has_impl(int))

    def test_get_impl(self):
        with self.assertRaises(DrbException):
            self.node[0].get_impl(str)



import unittest
import keyring

from unittest.mock import patch
from keyring.credentials import SimpleCredential
from requests.auth import HTTPBasicAuth

from drb.drivers.odata import ODataServiceNodeCSC
from drb.drivers.odata.odata_utils import ODataUtils
from drb.exceptions.odata import OdataRequestException
from tests.utils import start_mock_odata_csc, stop_mock_odata_csc


class TestAuthenticatedODataNode(unittest.TestCase):
    svc_url = None

    @classmethod
    def setUpClass(cls) -> None:
        cls.svc_url = 'http://auth.domain/csc'
        start_mock_odata_csc(cls.svc_url, auth_required=True)

    @classmethod
    def tearDownClass(cls) -> None:
        stop_mock_odata_csc()

    def test_auth_odata_node(self):
        auth = HTTPBasicAuth('foo', 'bar')
        odata = ODataServiceNodeCSC(self.svc_url, auth=auth)
        self.assertEqual(3, len(odata))

        # FIXME: odata.children[0] represents a Tar data and view as a TarNode
        #        due to the children auto resolution. The received TarBaseNode
        #        not really wrap its base node. So we change expected attribute
        node = odata.children[0]
        self.assertIn(('directory', None), node.attribute_names())
        self.assertEqual(True, node @ 'directory')
        self.assertEqual(True, node.get_attribute('directory'))

        with patch.object(ODataUtils, 'get_type_odata_svc',
                          side_effect=OdataRequestException()):
            with self.assertRaises(OdataRequestException):
                odata = ODataServiceNodeCSC(self.svc_url)
                node = odata.children[0]

    def test_auth_odata_node_with_key_ring(self):
        odata = ODataServiceNodeCSC(self.svc_url)
        with self.assertRaises(Exception):
            len(odata)

        auth = SimpleCredential('foo', 'bar')
        with patch.object(keyring, 'get_credential', return_value=auth):
            odata = ODataServiceNodeCSC(self.svc_url)
            self.assertEqual(3, len(odata))



import io
import unittest
import uuid

import unittest.mock as mocking

from drb.drivers.http.http import DrbHttpResponse
from drb.exceptions.core import DrbException, DrbNotImplementationException
from requests import Response

from drb.drivers.odata import ODataServiceNodeCSC, \
    ODataOrderNode, ODataProductNode
from drb.drivers.odata.odata_utils import ODataUtils
from tests.utils import start_mock_odata_csc, stop_mock_odata_csc


class TestODataOrders(unittest.TestCase):
    svc_url = 'https://gael-systems.com/odata/csc'
    orders = None
    service = None

    @classmethod
    def setUpClass(cls) -> None:
        start_mock_odata_csc(cls.svc_url)
        cls.service = ODataServiceNodeCSC(cls.svc_url)
        cls.orders = cls.service.orders

    @classmethod
    def tearDownClass(cls) -> None:
        stop_mock_odata_csc()

    def test_len_init(self):
        self.assertEqual(3, len(ODataServiceNodeCSC(self.svc_url).orders))

    def test_order_list(self):
        self.assertEqual(3, len(self.orders))
        self.assertEqual(
            self.orders[1:][0].name,
            self.orders[1].name)
        self.assertEqual(2, len(self.orders[0:2]))
        order_uuid = '56ac8297-e55e-4682-87c8-059e1e02260c'
        self.assertEqual(order_uuid, self.orders[uuid.UUID(order_uuid)].name)
        with self.assertRaises(KeyError):
            self.orders['Banana']

    def test_order_node(self):
        self.assertIsInstance(self.orders[0], ODataOrderNode)
        self.assertEqual('56ac8297-e55e-4682-87c8-059e1e02260c',
                         self.orders[0].name)
        self.assertIsInstance(self.orders[0].parent, ODataServiceNodeCSC)
        self.assertEqual('https://gael-systems.com/odata/csc',
                         self.orders[0].parent.name)
        self.assertEqual('https://gael-systems.com/odata/csc/'
                         'Orders(56ac8297-e55e-4682-87c8-059e1e02260c)',
                         self.orders[0].path.original_path)
        self.assertEqual([], self.orders[0].children)
        self.assertIsNone(self.orders[0].namespace_uri)
        self.assertIsNone(self.orders[0].value)
        self.assertFalse(self.orders[0].has_impl(io.BytesIO))
        self.assertFalse(self.orders[0].has_impl(str))
        with self.assertRaises(DrbNotImplementationException):
            self.assertIsNone(self.orders[0].get_impl(io.BytesIO))
        with self.assertRaises(DrbNotImplementationException):
            self.assertIsNone(self.orders[0].get_impl(str))

    def test_order_product(self):
        self.assertIsInstance(self.orders[0].product, ODataProductNode)
        self.assertEqual("S2B_OPER_MSI_L0__GR_MPS__20170803T123125"
                         "_S20170803T090229_D12_N02.05.tar",
                         self.orders[0].product.name)

    def test_order_attributes(self):
        self.assertIsInstance(self.orders[0].attributes, dict)
        # Attribute having None as value are not saved
        self.assertEqual(5, len(self.orders[0].attributes))
        self.assertEqual("in_progress", self.orders[0].get_attribute('Status'))

        with self.assertRaises(DrbException):
            self.orders[0].get_attribute('Banana')

        with self.assertRaises(DrbException):
            self.orders[0].get_attribute('Banana', 'Fruit')

    def test_order_cancel(self):
        with mocking.patch.object(ODataUtils, 'cancel_order',
                                  return_value=DrbHttpResponse(
                                      self.svc_url,
                                      Response())
                                  ):
            self.assertIsInstance(self.orders[0].cancel(), DrbHttpResponse)

    def test_order_by_uuid(self):
        self.assertIsInstance(
            self.orders[uuid.UUID('56ac8297-e55e-4682-87c8-059e1e02260c')],
            ODataOrderNode
        )

    def test_order_status(self):
        self.assertEqual(
            'in_progress',
            self.orders[uuid.UUID(
                '56ac8297-e55e-4682-87c8-059e1e02260c')].status
        )

    def test_order_eta(self):
        self.assertEqual(
            '2023-02-17T09:47:58.676Z',
            self.orders[uuid.UUID(
                '56ac8297-e55e-4682-87c8-059e1e02260c')].order_eta
        )

    def test_send_order(self):
        prd = self.service[uuid.UUID('0723ddbc-b0e7-4702-abeb-de257b9f4094')]
        self.assertIsInstance(prd.order(), ODataOrderNode)

        with mocking.patch.object(ODataProductNode, 'get_attribute',
                                  return_value=True
                                  ):
            with self.assertRaises(TypeError):
                prd.order()

        prd2 = self.service[uuid.UUID('0723d9a4-3bbe-305e-b712-5e820058e065')]
        with self.assertRaises(AttributeError):
            prd2.order()



import io
import unittest
import uuid
import re

import drb.drivers.tar
from drb.exceptions.core import DrbException

from drb.drivers.odata import ODataServiceNodeCSC, \
    ODataServiceType, ODataProductNode, ODataServiceNodeList, \
    ODataQueryPredicate
from tests.utils import start_mock_odata_csc, stop_mock_odata_csc


class TestODataServiceNode(unittest.TestCase):
    svc_url = 'https://gael-systems.com/odata/csc'
    node = None

    @classmethod
    def setUpClass(cls) -> None:
        start_mock_odata_csc(cls.svc_url)
        cls.node = ODataServiceNodeCSC(cls.svc_url)

    @classmethod
    def tearDownClass(cls) -> None:
        stop_mock_odata_csc()

    def test_type_service(self):
        self.assertEqual(ODataServiceType(1), self.node.type_service)
        self.node._type = None
        self.assertEqual(ODataServiceType(1), self.node.type_service)

    def test_name(self):
        self.assertEqual(self.svc_url, self.node.name)

    def test_namespace_uri(self):
        self.assertEqual('OData.CSC', self.node.namespace_uri)

    def test_value(self):
        self.assertIsNone(self.node.value)

    def test_path(self):
        self.assertEqual(self.svc_url, self.node.path.name)

    def test_parent(self):
        self.assertIsNone(self.node.parent)

    def test_attributes(self):
        self.assertEqual({}, self.node.attributes)

    def test_children(self):
        expected = [
            ODataProductNode(
                self.svc_url, '0723d9a4-3bbe-305e-b712-5e820058e065'),
            ODataProductNode(
                self.svc_url, '0723d9bf-02a2-3e99-b1b3-f6d81de84b62'),
            ODataProductNode(
                self.svc_url, '0723ddbc-b0e7-4702-abeb-de257b9f4094'),
        ]
        children = self.node.children
        self.assertIsNotNone(children)
        self.assertIsInstance(children, list)
        self.assertEqual(expected[1], children[1].get_impl(ODataProductNode))
        self.assertEqual(expected[-1], children[-2].get_impl(ODataProductNode))

        self.assertIsInstance(children[:2], ODataServiceNodeList)

        itr = iter(children)
        self.assertEqual(expected[0], next(itr).get_impl(ODataProductNode))
        self.assertEqual(expected[1], next(itr).get_impl(ODataProductNode))
        self.assertEqual(expected[2], next(itr).get_impl(ODataProductNode))

    def test_len(self):
        self.assertEqual(3, len(self.node))

    def test_get_attribute(self):
        with self.assertRaises(DrbException):
            att = self.node.get_attribute('foo')

    def test_has_impl(self):
        self.assertFalse(self.node.has_impl(io.BytesIO))

    def test_has_child(self):
        self.assertFalse(self.node.has_child('Banana'))
        self.assertTrue(self.node.has_child(
            'S2B_OPER_MSI_L0__GR_EPAE_20180703T214414_S201807'
            '03T165907_D05_N02.06.tar'
        ))
        self.assertFalse(self.node.has_child(
            'S2B_OPER_MSI_L0__GR_EPAE_20180703T214414_S2'
            '0180703T165907_D05_N02.06.tar',
            'Test'))

    def test_hash(self):
        self.assertEqual(hash(self.svc_url),
                         self.node.__hash__())

    def test_bracket_browse(self):
        prd2 = {
            'uuid': '0723d9bf-02a2-3e99-b1b3-f6d81de84b62',
            'name': 'S2B_OPER_MSI_L0__GR_EPAE_20180703T214414_S20180703T165907'
                    '_D05_N02.06.tar',
        }

        # FIXME: odata.children[0] represents a Tar data and view as a TarNode
        #        due to the children auto resolution. The received TarBaseNode
        #        not really wrap its base node. So we change expected attribute
        self.assertIsInstance(self.node[1], drb.drivers.tar.DrbBaseTarNode)
        self.assertEqual(self.node[1].name, prd2['name'])
        self.assertIsInstance(self.node[-3], drb.drivers.tar.DrbBaseTarNode)
        self.assertEqual(self.node[-3].name, prd2['name'])
        with self.assertRaises(IndexError):
            n = self.node[42]

        # tuple(str, str, int) -> by Name
        actual_node = self.node[prd2['name'], None, 0]
        self.assertEqual(actual_node.name, prd2['name'])
        actual_node = self.node[prd2['name'], 0]
        self.assertEqual(actual_node.name, prd2['name'])
        actual_node = self.node[prd2['name']]
        self.assertEqual(actual_node.name, prd2['name'])

        n = self.node[ODataQueryPredicate(filter=f"Name eq '{prd2['name']}'")]
        self.assertIsInstance(n, list)
        self.assertEqual(1, len(n))
        self.assertEqual(prd2['name'], n[0].name)

        with self.assertRaises(KeyError):
            n = self.node['test']
        with self.assertRaises(KeyError):
            n = self.node['test', 1]
        with self.assertRaises(KeyError):
            n = self.node[prd2['name'], 'ns']
        with self.assertRaises(KeyError):
            n = self.node[prd2['name'], 'ns', 1]

        # UUID -> by Id
        self.assertEqual(prd2['name'], self.node[uuid.UUID(prd2['uuid'])].name)
        with self.assertRaises(KeyError):
            fake_uuid = prd2['uuid'][:-2] + 'f'
            n = self.node[fake_uuid]

        with self.assertRaises(TypeError):
            n = self.node[b"helloWorld"]

    def test_equals(self):
        self.assertEqual(self.node, ODataServiceNodeCSC(self.svc_url))

    def test_prepare_filter(self):
        pat = r"\d{4}-\d{2}-\d{2}\w{3}\W\w{2}\W\w{3}"

        self.assertTrue(re.search(pat,
                                  self.node.prepare_filter(None)))
        self.assertIn('PublicationDate', self.node.prepare_filter(None))
        self.assertEqual('PublicationDate',
                         self.node.prepare_filter('PublicationDate'))
        self.assertIn('foo', self.node.prepare_filter('foo'))



from drb.drivers.odata.factory import OdataFactory
from drb.drivers.odata.odata_nodes import (
    ODataAttributeNode, ODataProductNode, ODataOrderNode
)
from drb.drivers.odata.odata_services_nodes import (
    ODataServiceNodeList,
    ODataServiceNodeCSC,
    ODataServiceNodeDias,
    ODataServiceNodeDhus
)

from drb.drivers.odata.expression import (
    Expression, BinaryExpr, FunctionExpr, NotExpr,
    GroupExpr, PropExpr, BooleanExpr,
    NumberExpr, StringExpr, ExpressionFunc,
    ExpressionType, LogicalOperator, ComparisonOperator,
    Footprint, ArithmeticOperator, GroupingOperator)
from drb.drivers.odata.odata_utils import ODataServiceType, ODataQueryPredicate
from . import _version

__version__ = _version.get_versions()['version']

__all__ = [
    'ODataServiceNodeList',
    'ODataServiceNodeCSC',
    'ODataServiceNodeDias',
    'ODataServiceNodeDhus',
    'ODataProductNode',
    'ODataAttributeNode',
    'ODataQueryPredicate',
    'ODataServiceType',
    'ODataOrderNode',
    'OdataFactory',
    'ExpressionFunc',
    'ExpressionType',
    'LogicalOperator',
    'ComparisonOperator',
    'ArithmeticOperator',
    'GroupingOperator'
]



import json
import os
import unittest

from drb.exceptions.core import DrbException

from drb.drivers.odata import ODataProductNode, ODataServiceNodeCSC
from tests.utils import start_mock_odata_csc, stop_mock_odata_csc


class TestProductNode(unittest.TestCase):
    service_url = 'https://domain.com/csc'
    uuid = '0723d9a4-3bbe-305e-b712-5e820058e065'
    name = 'S2B_OPER_MSI_L0__GR_MPS__20170803T123125_S20170803T090229_' \
           'D12_N02.05.tar'
    node = None

    @classmethod
    def setUpClass(cls) -> None:
        start_mock_odata_csc(cls.service_url)
        cls.node = ODataProductNode(cls.service_url, cls.uuid)

    @classmethod
    def tearDownClass(cls) -> None:
        stop_mock_odata_csc()

    def test_name(self):
        self.assertEqual(self.name, self.node.name)

    def test_namespace_uri(self):
        self.assertIsNone(self.node.namespace_uri)

    def test_value(self):
        self.assertIsNone(self.node.value)

    def test_path(self):
        self.assertEqual(f'{self.service_url}/Products({self.uuid})',
                         self.node.path.name)

    def test_parent(self):
        self.assertIsNone(self.node.parent)
        parent = ODataServiceNodeCSC(self.service_url)
        prd_node = parent[0]
        self.assertEqual(parent, prd_node.parent)

    def test_attributes(self):
        node = self.node.get_impl(ODataProductNode)
        attributes = node.attributes
        path = os.path.join(os.path.dirname(__file__), 'resources',
                            'product1.json')
        with open(path) as f:
            data = json.load(f)

        name = 'Id'
        self.assertEqual(self.uuid, node @ name)
        self.assertEqual(self.uuid, attributes[name, None])
        self.assertEqual(self.uuid, self.node.get_attribute(name))

        name = 'Name'
        self.assertEqual(self.name, attributes[name, None])
        self.assertEqual(self.name, self.node.get_attribute(name))

        name = 'ContentType'
        self.assertEqual(data[name], attributes[name, None])
        self.assertEqual(data[name], self.node.get_attribute(name))

        name = 'ContentLength'
        self.assertEqual(data[name], attributes[name, None])
        self.assertEqual(data[name], self.node.get_attribute(name))

        name = 'OriginDate'
        self.assertEqual(data[name], attributes[name, None])
        self.assertEqual(data[name], self.node.get_attribute(name))

        name = 'PublicationDate'
        self.assertEqual(data[name], attributes[name, None])
        self.assertEqual(data[name], self.node.get_attribute(name))

        name = 'ModificationDate'
        self.assertEqual(data[name], attributes[name, None])
        self.assertEqual(data[name], self.node.get_attribute(name))

        name = 'EvictionDate'
        self.assertEqual(data[name], attributes[name, None])
        self.assertEqual(data[name], self.node.get_attribute(name))

        name = 'Online'
        self.assertTrue(attributes[name, None])
        self.assertTrue(self.node.get_attribute(name))

        name = 'Checksum'
        self.assertEqual(data[name], attributes[name, None])
        self.assertEqual(data[name], self.node.get_attribute(name))

        name = 'ContentDate'
        self.assertEqual(data[name], attributes[name, None])
        self.assertEqual(data[name], self.node.get_attribute(name))

        name = 'Footprint'
        # FIXME: Attributes having null as value are not take in consideration
        self.assertNotIn((name,  None), self.node.attribute_names())
        # self.assertIsNone(attributes[name, None])
        # self.assertIsNone(self.node.get_attribute(name))

        with self.assertRaises(KeyError):
            attributes['@odata.context', None]
        with self.assertRaises(DrbException):
            self.node.get_attribute('@odata.context')

    def test_children(self):
        children = self.node.get_impl(ODataProductNode).children
        self.assertIsNotNone(children)
        self.assertTrue(self.node.has_child("Attributes"))
        self.assertFalse(self.node.has_child("Attributes", 'ns'))
        self.assertIsInstance(children, list)
        self.assertEqual(1, len(children))
        self.assertEqual('Attributes', children[0].name)
